# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import multiprocessing as mp
import os
from math import trunc
from typing import cast, Tuple

import numpy as np
import pandas as pd

from rinpy import utils
from rinpy.constants import *
from rinpy.log_messages import LOG_MESSAGE
from rinpy.log_util import log_elapsed_time
from rinpy.utils import create_folder_not_exists, PdbRecord, clear_existing_file, write_ppdb_to_pdb_file, \
    create_empty_ppdb_atom_df, create_ppdb_atom_df

SOURCE = 'source'
TARGET = 'target'
AFFINITY = 'affinity'


class ResidueNetworkBuilder:
    """ The Residue Network Builder module generates output data for constructing residue interaction networks based on
        affinity calculations between α-carbon atoms. It computes atom-to-atom distances between source and target
        residues.

        Attributes
        ----------
        pdb_name : str, default: None
            PDB ID which must be same with file name, such as 4OBE (4OBE.pdb).

        pdb_path : str, default: None
            Location of the PDB file that will be used for contact calculation (path/to/4OBE.pdb).

        use_preprocess : bool, default: False
            Whether to preprocess the PDB before analysis. The given PDB file contains information
            such as {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'}.If this is True, then only ATOM records from
            the PDB file will be extracted as preprocess and will be saved as 4OBE_pre.pdb in the given output of 4OBE.

        cutoff : float, default: 4.5 Armstrong (Å)
            The distance between atoms will be considered if it is equal or less than the given cutoff.

        destination_output_path : str, default: output, this path must be provided.
            Location of the output results in which all files for each PDB file process are stored in this path.

        het_atom_list: list, default: None
            List of hetero atoms to be included as a node in the generated graph.

        remove_hydrogen: bool, default: False
            Remove the hydrogen atoms from the given PDB file.

        num_workers: int, default: None The number of CPU cores will be used for the heavy atom-atom calculation.
        If None, available CPU cores will be used.
    """

    def __init__(self, pdb_name: str = None,
                 pdb_path: str = None,
                 use_preprocess: bool = False,
                 cutoff: float = 4.5,
                 destination_output_path: str = None,
                 het_atom_list: list = None,
                 remove_hydrogen: bool = True,
                 num_workers=None):
        if destination_output_path is None:
            raise ValueError('You must provide an output path to proceed.')
        self.pdb_name = pdb_name
        self.pdb_path = pdb_path
        self.het_atom_list = het_atom_list
        self.remove_hydrogen = remove_hydrogen
        self.ppdb = utils.get_ppdb(pdb_file_path=self.pdb_path)
        self.cutoff = cutoff
        logging.info(f"Cutoff: {self.cutoff}")
        self.num_workers = num_workers
        workers_msg = (str(self.num_workers) if self.num_workers is not None else "Detected CPU cores will be used.")
        logging.info(f"CPU workers requested: {workers_msg}")
        self.destination_output_path = destination_output_path
        self.actual_residue_number_map = dict()

        if use_preprocess:
            self._preprocess()

        if het_atom_list is not None:
            self._append_het_atoms()

        self.output_path = os.path.join(self.destination_output_path, self.pdb_name)

        create_folder_not_exists(self.output_path)

        self.residue_edges_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                                    f'{self.pdb_name}_{RESIDUE_EDGES_FILE}')

        self.residue_average_coordinates_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                                                  f'{self.pdb_name}_{RESIDUE_AVERAGE_COORDINATES_FILE}')

        self.residue_average_pdb_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                                          f'{self.pdb_name}_{RESIDUE_AVERAGE_PDB_FILE}')

    def get_number_of_amino_acid(self):
        """ Returns the total number of amino acid for the given PDB file """
        return self._get_pdb_df()[PdbRecord.ATOM.name].groupby(
            [CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]).ngroups

    def get_actual_residue_number_map(self):
        """ Returns the actual residue map for the given PDB file. This file contains Chain ID and residue number
        corresponding residue index.
        """
        return self.actual_residue_number_map

    def _get_pdb_df(self):
        """ Returns dictionary storing pandas DataFrames for PDB record section.
        The dictionary keys are {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'}
        """
        return self.ppdb.df

    def _preprocess(self):
        """ Only gets the ATOM record section from the dictionary that stores {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'}"""
        if self.remove_hydrogen:
            logging.info(f'Removing hydrogen...')
            self.ppdb.df[PdbRecord.ATOM.name] = self.ppdb.df[PdbRecord.ATOM.name][
                self.ppdb.df[PdbRecord.ATOM.name][ELEMENT_SYMBOL] != HYDROGEN_SYMBOL]

            self.ppdb.to_pdb(
                path=os.path.join(os.path.dirname(self.pdb_path), f'{self.pdb_name}{PREPROCESS_SAVE_EXT}'),
                records=[PdbRecord.ATOM.name],
                gz=False,
                append_newline=True)

            logging.info(f"{self.pdb_name}{PREPROCESS_SAVE_EXT} {LOG_MESSAGE['PREPROCESS_SAVED']}")

    def _append_het_atoms(self):
        """ Append HETERO ATOMS to the ATOM section of the pdb file which will be included in the graph generated."""
        if self.het_atom_list is not None:
            logging.info(f'Appending HETERO ATOMS ...')
            self.ppdb.df[PdbRecord.ATOM.name] = pd.concat(
                [self.ppdb.df[PdbRecord.ATOM.name], self._get_selected_het_atoms()], ignore_index=True)

            self.ppdb.to_pdb(
                path=os.path.join(os.path.dirname(self.pdb_path), f'{self.pdb_name}{WITH_HET_ATOM_SAVE_EXT}'),
                records=[PdbRecord.ATOM.name],
                gz=False,
                append_newline=True)

            logging.info(f"{self.pdb_name}{WITH_HET_ATOM_SAVE_EXT} {LOG_MESSAGE['PREPROCESS_SAVED']}")

    def clear_files(self):
        """ Clears created and stored files for the given PDB file."""
        clear_existing_file(file_path=self.residue_average_pdb_file_path)
        clear_existing_file(file_path=self.residue_average_coordinates_file_path)
        clear_existing_file(file_path=self.residue_edges_file_path)

    def _get_selected_het_atoms(self):
        """ het_atom is tuple which contains (chain_id, residue_number, residue_name, insertion) """
        matching_rows_df_list = []
        if self.het_atom_list is not None:
            het_atom_df = self._get_pdb_df()[PdbRecord.HETATM.name]
            for het_atom in self.het_atom_list:
                matched_rows = het_atom_df[(het_atom_df[CHAIN_ID] == het_atom[0]) &
                                           (het_atom_df[RESIDUE_NUMBER] == het_atom[1]) &
                                           (het_atom_df[RESIDUE_NAME] == het_atom[2]) &
                                           (het_atom_df[INSERTION] == het_atom[3])]
                if not matched_rows.empty:
                    matching_rows_df_list.append(matched_rows)

        matching_rows_df = pd.concat(matching_rows_df_list,
                                     ignore_index=True) if matching_rows_df_list else pd.DataFrame(columns=PDB_COLUMNS)
        return matching_rows_df

    def _calculate_affinity(self, index_pair, residue_number_list, residue_coord_dict):
        """ Calculates the local interaction strength between two residues based on the spatial distance
        between their atom coordinates.

        Parameters
        ----------
        index_pair : tuple[int, int]
            A tuple containing the indices (i, j) of the source and target residues within the residue_number_list.

        residue_number_list : list[tuple[str, int, str]]
            A list of residue identifiers where each item is a tuple of (chain ID, residue number, insertion code).

        residue_coord_dict : dict[tuple[str, int, str], list[dict[str, float]]]
            A dictionary mapping residue identifiers to their 3D coordinates.
            Each value is a list containing one dictionary with keys: 'x_coord', 'y_coord', and 'z_coord'.

        Returns
        -------
        dict or None
            A dictionary containing the 1-based indices of the source and target residues
            and the calculated affinity value, or None if the affinity could not be computed.

            Example:
            {
                'source': 1,
                'target': 2,
                'affinity': 0.12345678
            }
        """
        i, j = index_pair
        source_list = residue_coord_dict[residue_number_list[i]]
        target_list = residue_coord_dict[residue_number_list[j]]
        if not source_list or not target_list:
            return None
        source_coords = np.array([(source[X_COORD], source[Y_COORD], source[Z_COORD]) for source in source_list])
        target_coords = np.array([(target[X_COORD], target[Y_COORD], target[Z_COORD]) for target in target_list])
        distances = np.linalg.norm(source_coords[:, np.newaxis] - target_coords, axis=2)
        affinity_count = np.sum(distances <= self.cutoff)
        if affinity_count != 0:
            affinity = affinity_count / (np.sqrt(len(source_list)) * np.sqrt(len(target_list)))
            return {
                SOURCE: i + 1,
                TARGET: j + 1,
                AFFINITY: trunc(affinity * np.power(10, 8)) / np.power(10, 8)
            }
        else:
            return None

    @staticmethod
    def _get_default_atom_name(residue_name):
        if residue_name in RNA_BASES or residue_name in DNA_BASES:
            return P_ATOM_NAME
        return CA_ATOM_NAME

    @staticmethod
    def _get_record_name(res_atom_df, validate=False):
        """ returns 'ATOM' or 'HETATM' for a residue-level atom dataframe.
            Assumes all atoms in the residue share the same record_name.
        """
        if validate:
            if res_atom_df[RECORD_NAME].nunique() != 1:
                raise ValueError(f"Mixed {ATOM}/{HETATM} found in the same residue.")

        return res_atom_df[RECORD_NAME].iloc[0]

    def calculate_contact(self):
        """Calculate the atom-atom distance between residues."""
        residue_average_pdb_list = []
        residue_average_coords_list = []
        atoms_df = self._get_pdb_df()[PdbRecord.ATOM.name]

        residue_chain_dict = {
            (str(residue_name), str(chain_id), int(residue_number), str(insertion), str(segment_id)): rows for
            (residue_name, chain_id, residue_number, insertion, segment_id), rows in
            atoms_df.groupby([RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID], sort=False)}

        residue_coord_dict = {}
        residue_number_list = []

        default_atom_names = set()
        residue_index = 1
        for res_num_key, res_atom_df in residue_chain_dict.items():
            residue_number_list.append(res_num_key)
            average_values = res_atom_df[[X_COORD, Y_COORD, Z_COORD, B_FACTOR]].mean()
            x_coord_average = average_values[X_COORD]
            y_coord_average = average_values[Y_COORD]
            z_coord_average = average_values[Z_COORD]
            b_factor_average = average_values[B_FACTOR]
            temp_coord_list = []

            for _, row in res_atom_df.iterrows():
                temp_coord_list.append({X_COORD: row[X_COORD],
                                        Y_COORD: row[Y_COORD],
                                        Z_COORD: row[Z_COORD]
                                        })

            residue_coord_dict[res_num_key] = temp_coord_list

            residue_name, chain_id, residue_number, insertion, segment_id = cast(Tuple[str, str, int, str, str],
                                                                                 res_num_key)
            atom_name = self._get_default_atom_name(residue_name=residue_name)
            default_atom_names.add(atom_name)

            self.actual_residue_number_map[residue_index] = res_num_key + (self._get_record_name(res_atom_df),)

            residue_average_pdb_list.append({RECORD_NAME: PdbRecord.ATOM.name,
                                             ATOM_NUMBER: residue_index,
                                             ATOM_NAME: atom_name,
                                             RESIDUE_NAME: residue_name,
                                             CHAIN_ID: chain_id,
                                             RESIDUE_NUMBER: residue_number,
                                             INSERTION: insertion,
                                             SEGMENT_ID: segment_id,
                                             X_COORD: x_coord_average,
                                             Y_COORD: y_coord_average,
                                             Z_COORD: z_coord_average,
                                             B_FACTOR: b_factor_average
                                             })
            residue_average_coords_list.append({RESIDUE_NUMBER: residue_index,
                                                X: THREE_DECIMAL_FLOAT_FORMATTER.format(x_coord_average),
                                                Y: THREE_DECIMAL_FLOAT_FORMATTER.format(y_coord_average),
                                                Z: THREE_DECIMAL_FLOAT_FORMATTER.format(z_coord_average)
                                                })

            residue_index += 1

        logging.info(f'Default atom name(s) for {self.pdb_name} are [{", ".join(sorted(default_atom_names))}].')

        residue_average_coords_df = pd.DataFrame(residue_average_coords_list)

        residue_average_coords_df.to_csv(
            path_or_buf=str(self.residue_average_coordinates_file_path),
            sep='\t',
            index=False,
            header=False)

        write_ppdb_to_pdb_file(create_empty_ppdb_atom_df(create_ppdb_atom_df(pd.DataFrame(residue_average_pdb_list))),
                               self.residue_average_pdb_file_path)

        residue_edges_list = self._process_affinity_calculation(residue_number_list, residue_coord_dict)
        if residue_edges_list:
            residue_edges_df = pd.DataFrame(residue_edges_list, columns=[SOURCE, TARGET, AFFINITY])

            residue_edges_df.to_csv(str(self.residue_edges_file_path),
                                    sep='\t',
                                    index=False,
                                    header=False)
            return True
        else:
            return False

    @log_elapsed_time(message="CALCULATE AFFINITY")
    def _process_affinity_calculation(self, residue_number_list, residue_coord_dict):
        """ Calculates all affinity calculation for residues.

        Parameters
        ----------
        residue_number_list : list[tuple[str, int, str]]
            A list of residue identifiers where each item is a tuple of (chain ID, residue number, insertion code).

        residue_coord_dict : dict[tuple[str, int, str], list[dict[str, float]]]
            A dictionary mapping residue identifiers to their 3D coordinates.
            Each value is a list containing one dictionary with keys: 'x_coord', 'y_coord', and 'z_coord'.

        Returns
        -------
        list of dict
            List of a dictionary containing the 1-based indices of the source and target residues
            and the calculated affinity value.

            Example:
            [{
                'source': 1,
                'target': 2,
                'affinity': 0.12345678
            },
            {
                'source': 1,
                'target': 3,
                'affinity': 2.1558478
            },
            ]
        """

        total_residue_number = len(residue_number_list)
        index_pairs = [(i, j) for i in range(total_residue_number - 1) for j in range(i + 1, total_residue_number)]
        cpu_count = mp.cpu_count()
        if self.num_workers is not None:
            cpu_count = self.num_workers
        logging.info(f"CPU cores detected: {mp.cpu_count()} | CPU workers requested: {cpu_count}")
        with mp.Pool(cpu_count) as pool:
            results = pool.starmap(self._calculate_affinity,
                                   [(index_pair, residue_number_list, residue_coord_dict) for index_pair in
                                    index_pairs])

        residue_edges_list = [result for result in results if result is not None]
        residue_edges_list = sorted(residue_edges_list, key=lambda x: (x[SOURCE], x[TARGET]))
        return residue_edges_list


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#

__version__ = "0.2.4"
__author__ = "Zehra Sarica <zehraacar559@gmail.com, sarica16@itu.edu.tr>"

import rinpy.logging_config
from rinpy.network_comparator import NetworkComparator
from rinpy.rin_process import RINProcess

__all__ = ["RINProcess", "NetworkComparator"]



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os

from rinpy.constants import CHAIN_ID, RESIDUE_NUMBER, INSERTION, RESIDUE_NAME, SEGMENT_ID
from rinpy.utils import CentralityType

_PML = "pml"
_RESIDUES = "residues"


class PymolUtils:
    def __init__(self, pdb_name: str, destination_output_path: str,
                 actual_residue_number_map: dict[int, tuple[str, str, int, str, str, str]]):
        self.pdb_name = pdb_name
        self.destination_output_path = destination_output_path
        self.actual_residue_number_map = actual_residue_number_map

    def get_full_save_path(self, filename: str, extension: str = "png"):
        return os.path.join(self.destination_output_path, self.pdb_name, f"{self.pdb_name}_{filename}.{extension}")

    def generate_pymol_query_string(self, out_df, centrality_type: CentralityType):

        full_path = self.get_full_save_path(filename=f'{centrality_type.display_name}_pymol_queries', extension="txt")

        query_str_df = (out_df.groupby(CHAIN_ID)[[RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]]
                        .apply(lambda df: list(df.itertuples(index=False, name=None))).reset_index(name=_RESIDUES))

        query_strings = []

        for _, row in query_str_df.iterrows():
            residues = row[_RESIDUES]
            formatted_residues = '+'.join(self.generate_selected_residues_by_tuple(residues))
            query = f"select {self.pdb_name} and {formatted_residues}"
            query_strings.append(query)

        with open(full_path, "w") as file:
            for query in query_strings:
                file.write(query + "\n")

        logging.info(f'PyMOL query string has been saved to {full_path}.')

    def export_pymol_script(self, full_path_to_pdb, residues, filename=None, is_actual_residue=False, full_path=None):
        """ Generate a PyMOL script to visualize residues.

        Parameters:
        - residues: List of residue numbers (int)
        """
        if full_path is None:
            full_path = self.get_full_save_path(filename=filename, extension=_PML)

        with open(full_path, 'w') as file:
            file.write(f"load ./{os.path.basename(full_path_to_pdb)}\n")
            file.write("hide everything\n")
            file.write("show cartoon\n")
            file.write("color gray80\n\n")
            if is_actual_residue:
                selected_residues = "+".join(self._generate_selected_high_percentage_residues(residues))
            else:
                selected_residues = " + ".join(self._generate_selected_residues(residues))

            file.write(f"select residues, {selected_residues}\n")
            file.write("show cartoon, residues\n")
            file.write("color warmpink, residues\n")
            file.write("set stick_radius, 0.25, residues\n\n")

            file.write("set label_size, 16\n")
            file.write("set label_font_id, 7\n")
            file.write("label residues and name CA, resn + resi\n")
            file.write("set label_position, [0.0, -0.3, -0.5], residues\n")
            file.write("set label_color, white\n\n")

            file.write("\nzoom residues, 10\n")
            file.write("deselect\n")

    @staticmethod
    def _format_resi(res_num, ins_code):
        has_ins = ins_code not in ["", "''", None]
        return f"{res_num}{ins_code}" if has_ins else str(res_num)

    def _generate_selected_residues(self, residues):
        selected_residue_list = []
        for residue_index in residues:
            if (res_info := self.actual_residue_number_map.get(residue_index)) is not None:
                res_name, chain_id, res_num, ins_code, seg_id, record_name = res_info
                resi_str = self._format_resi(res_num, ins_code)
                seg_str = f" and segi {seg_id}" if seg_id not in ["", "''", None] else ""
                sel = f"(chain {chain_id} and resi {resi_str} and resn {res_name}{seg_str})"
                selected_residue_list.append(sel)
        return selected_residue_list

    def generate_selected_residues_by_tuple(self, residues):
        selected_residue_list = []
        for res_name, chain_id, res_num, ins_code, seg_id in residues:
            resi_str = self._format_resi(res_num, ins_code)
            seg_id_str = f" and segi {seg_id}" if seg_id not in ["", "''", None] else ""
            sel = f"(chain {chain_id} and resi {resi_str} and resn {res_name}{seg_id_str})"
            selected_residue_list.append(sel)
        return selected_residue_list

    @staticmethod
    def _generate_selected_high_percentage_residues(high_percentage_residues):
        selected_high_percentage_residues_list = []
        for res_name, chain_id, res_num, ins_code, seg_id in high_percentage_residues:
            has_ins = ins_code not in [None, '', "''"]
            resi_str = f"{res_num}{ins_code}" if has_ins else str(res_num)
            seg_str = f" and segi {seg_id}" if seg_id not in [None, '', "''"] else ""
            sel = f"(chain {chain_id} and resi {resi_str} and resn {res_name}{seg_str})"
            selected_high_percentage_residues_list.append(sel)
        return selected_high_percentage_residues_list

    def export_pymol_script_hinge(self, full_path_to_pdb, residues, filename=None, is_actual_residue=False,
                                  full_path=None, high_percentage_residues=None):
        """ Generate a PyMOL script to visualize residues.

        Parameters:
        - residues: List of residue numbers (int)
        """
        if full_path is None:
            full_path = self.get_full_save_path(filename=filename, extension="pml")

        with open(full_path, 'w') as file:
            file.write(f"load ./{os.path.basename(full_path_to_pdb)}\n")
            file.write("hide everything\n")
            file.write("show cartoon\n")
            file.write("color gray80\n\n")
            if is_actual_residue:
                selected_residues = "+".join(str(residue_index) for residue_index in residues)
            else:
                selected_residues = " + ".join(self._generate_selected_residues(residues))

            file.write(f"select hinge_residues, {selected_residues}\n")
            file.write("show cartoon, hinge_residues\n")
            file.write("color yellow, hinge_residues\n")
            file.write("set stick_radius, 0.25, hinge_residues\n\n")

            if high_percentage_residues is not None:
                selected_high_percentage_residues = "+".join(
                    self._generate_selected_high_percentage_residues(high_percentage_residues))
                file.write(f"select hub_residues, {selected_high_percentage_residues}\n")
                file.write("show cartoon, hub_residues\n")
                file.write("color warmpink, hub_residues\n")
                file.write("set stick_radius, 0.25, hub_residues\n\n")

            file.write("set label_size, 16\n")
            file.write("set label_font_id, 7\n")

            file.write("label hinge_residues and name CA, resn + resi\n")
            file.write("set label_position, [0.0, -0.3, -0.5], hinge_residues\n")

            if high_percentage_residues is not None:
                file.write("label hub_residues and name CA, resn + resi\n")
                file.write("set label_position, [0.0, -0.3, -0.5], hub_residues\n")

            file.write("set label_color, white\n\n")

            file.write("\nzoom hinge_residues or hub_residues, 10\n")
            file.write("deselect\n")


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

from rinpy.utils import *

_END = "END"
_ATOM = "ATOM"
_OCCUPANCY = 'occupancy'


class Trajectory:

    def __init__(self, input_trajectory_file, output_path, stride=10):
        if output_path is None:
            raise ValueError('You must provide an output path to proceed.')

        self.input_trajectory_file = input_trajectory_file
        self.output_path = output_path
        self.stride = stride

    def parse_to_pdb(self):
        """ All extracted PDBs absolute path list
        :return: return list: ['output/0001/0001.pdb', 'output/0002/0002.pdb', 'output/0003/0003.pdb']
        """
        delete_folders_in_path(self.output_path)

        snapshots = []
        current_snapshot = []

        lines = self.read_lines()

        for line in lines:
            if line.startswith(_END):
                snapshots.append(current_snapshot)
                current_snapshot = []
            elif line.startswith(_ATOM):
                current_snapshot.append(line)

        num_snapshots = len(snapshots)
        output_path_pdb_ids = []

        for i in range(0, num_snapshots, self.stride):
            snapshot = snapshots[i]
            filename = f"{i + 1:04d}"
            parent_dir = os.path.join(self.output_path, filename)
            create_folder_not_exists(parent_dir)
            output_file = os.path.join(parent_dir, f"{filename}.pdb")

            ppdb = PandasPdb()
            ppdb.read_pdb_from_list(snapshot)
            ppdb.df[_ATOM].loc[ppdb.df[_ATOM][
                _OCCUPANCY].isna(), _OCCUPANCY] = 0  # if occupancy is NAN value, 0 value is set as default.
            write_ppdb_to_pdb_file(ppdb, output_file)
            output_path_pdb_ids.append(output_file)
            logging.info(f"Snapshot {i + 1} saved as '{output_file}'")

        logging.info(f"{output_path_pdb_ids}")
        return output_path_pdb_ids

    def read_lines(self):
        with open(self.input_trajectory_file, 'r') as file:
            lines = file.readlines()
        return lines


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""
import functools
import logging
import time

from rinpy.constants import STAR_PRINT_COUNT


def log_time(message=None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            msg_prefix = f"{message} " if message else "x"
            logging.info(f"{msg_prefix} started at: {time.ctime(start_time)}")
            result = func(*args, **kwargs)
            end_time = time.time()
            logging.info(f"{msg_prefix} ended at: {time.ctime(end_time)}")
            duration = end_time - start_time
            logging.info(f"{msg_prefix}Duration: {duration:.2f} seconds")
            return result

        return wrapper

    return decorator


def log_details(message):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            msg = message(*args, **kwargs) if callable(message) else message
            logging.info(f"{msg} STARTS...")
            result = func(*args, **kwargs)
            logging.info(f"{msg} ENDS...")
            return result

        return wrapper

    return decorator


def log_star_message(message: str, star_count: int = STAR_PRINT_COUNT):
    stars = "*" * star_count
    logging.info(f'{stars} {message} {stars}')


def log_with_stars(label: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            log_star_message(f"{label.upper()} STARTS")
            result = func(*args, **kwargs)
            log_star_message(f"{label.upper()} ENDS")
            return result

        return wrapper

    return decorator


def log_elapsed_time_detail(msg, start, end):
    duration = end - start
    logging.info(f"{msg} is {duration:.2f} seconds.")


def log_elapsed_time(message: str = ""):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self_obj = args[0]  # first argument to a method is always 'self'
            pdb_name = getattr(self_obj, "pdb_name", "UnknownPDB")
            start_time = time.time()
            logging.info(f'{message} STARTs at {time.ctime(start_time)}')
            result = func(*args, **kwargs)
            end_time = time.time()
            logging.info(f'{message} ENDs at {time.ctime(end_time)}')
            duration = end_time - start_time
            logging.info(f"Elapsed duration of {message} for {pdb_name.upper()} is {duration:.2f} seconds.")
            return result

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os

import pandas as pd
from biopandas.pdb import PandasPdb

from rinpy import utils
from rinpy.constants import *
from rinpy.log_util import log_time
from rinpy.utils import create_folder_not_exists, PdbRecord, write_ppdb_to_pdb_file, CentralityType

PDB_FILE = 'pdb_file'
OUTPUT_FILE = 'output_file'
KEY = 'key'


class CentralityPdbMapper:
    """ Creates centrality hub PDBs by replacing b-factor value with centrality score of the given centrality type.
    Here, the given pdb is regenerated by replacing b-factor value with centrality score of each centrality type
    This allows to visualize centrality scores on the pdb via PyMol or other tools supporting pdb format.

    Attributes
    ----------
    pdb_name : str, default: None
        PDB ID which must be same with file name, such as 4OBE (4OBE.pdb).
    destination_output_path : str, default: output, this path must be provided.
        Location of the output results in which all files for each PDB file process are stored in this path.
    calculation_options: dict, default: None
        The calculation options as the input for the whole process.
        Example:
        {
            'remove_hydrogen': {
                'is_checked': True,
                'value': 0
            },
            'betweenness': {
                'is_checked': True,
                'value': 5
            },
            'closeness': {
                'is_checked': False,
                'value': 10
            },
            'degree': {
                'is_checked': False,
                'value': 10
            },
            'cluster_number': {
                'is_checked': None,
                'value': 4
            },
            'cutoff': {
                'is_checked': True,
                'value': 4.5
            },
        }
        """

    def __init__(self, graph, pdb_name=None, destination_output_path=None, calculation_options=None):
        if destination_output_path is None:
            raise ValueError('You must provide an output path to proceed.')
        self.graph = graph
        self.pdb_name = pdb_name
        self.destination_output_path = destination_output_path
        create_folder_not_exists(self.destination_output_path)
        self.residue_average_pdb_file_path = os.path.join(
            os.path.join(str(self.destination_output_path), self.pdb_name),
            f'{self.pdb_name}_{RESIDUE_AVERAGE_PDB_FILE}')
        self.calculation_options = calculation_options
        self.ppdb = utils.get_ppdb(pdb_file_path=self.residue_average_pdb_file_path)

    @log_time("Pdb Mapper Process")
    def process(self):
        logging.info(f'Centrality PDB Mapper has been started...')
        base_file_path = os.path.join(self.destination_output_path, self.pdb_name)
        columns = [ATOM_NUMBER, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]
        dtypes = {ATOM_NUMBER: int, CENTRALITY_SCORE: float, RESIDUE_NAME: str, CHAIN_ID: str, RESIDUE_NUMBER: int,
                  INSERTION: str, SEGMENT_ID: str}
        sort_keys = [ATOM_NUMBER]

        centrality_types = {
            CentralityType.BET.display_name: {
                OUTPUT_FILE: CENTRALITY_CSV_TEMPLATE.format(type=CentralityType.BET.display_name),
                PDB_FILE: CENTRALITY_PDB_TEMPLATE.format(type=CentralityType.BET.display_name)
            },
            CentralityType.CLOS.display_name: {
                OUTPUT_FILE: CENTRALITY_CSV_TEMPLATE.format(type=CentralityType.CLOS.display_name),
                PDB_FILE: CENTRALITY_PDB_TEMPLATE.format(type=CentralityType.CLOS.display_name)
            },
            CentralityType.DEG.display_name: {
                OUTPUT_FILE: CENTRALITY_CSV_TEMPLATE.format(type=CentralityType.DEG.display_name),
                PDB_FILE: CENTRALITY_PDB_TEMPLATE.format(type=CentralityType.DEG.display_name)
            }
        }

        for centrality, files in centrality_types.items():
            option = self.calculation_options.get(centrality, {})
            if not option.get(IS_CHECKED, False):
                continue

            output_file = os.path.join(str(base_file_path), f"{self.pdb_name}_{files[OUTPUT_FILE]}")
            df = utils.get_df(output_file_path=output_file,
                              columns=columns,
                              dtypes=dtypes,
                              sort_keys=sort_keys)
            hub_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                         f"{self.pdb_name}_{files[PDB_FILE]}")
            self._write_to_pdb(df, hub_file_path)
            self._write_to_original_pdb_file(df, files[PDB_FILE])
        logging.info(f'Centrality Hub analyses has been done...')

    def _write_to_pdb(self, out_df, out_filename):
        logging.info(f'Starting creation of {out_filename}.')

        atom_df = utils.get_atom_pdb_df(ppdb=self.ppdb).copy()
        out_df = out_df.copy()

        out_df[ATOM_NUMBER] = out_df[ATOM_NUMBER].astype(int)
        atom_df[ATOM_NUMBER] = atom_df[ATOM_NUMBER].astype(int)

        merged_df = atom_df.merge(out_df[[ATOM_NUMBER, CENTRALITY_SCORE]], on=ATOM_NUMBER, how='left')
        merged_df[B_FACTOR] = merged_df[CENTRALITY_SCORE].apply(
            lambda x: float(TWO_DECIMAL_FLOAT_FORMATTER.format(x)) if pd.notna(x) else merged_df[B_FACTOR])

        merged_df[BLANK_4] = ''
        merged_df[OCCUPANCY] = merged_df[OCCUPANCY].fillna(0)

        clean_df = merged_df.copy()
        for col in [CENTRALITY_SCORE]:
            if col in clean_df.columns:
                clean_df = clean_df.drop(columns=[col])

        ppdb = PandasPdb()
        ppdb.df[PdbRecord.ATOM.name] = clean_df
        write_ppdb_to_pdb_file(ppdb=ppdb, out_filename=out_filename)

        logging.info(f'Finished creating {out_filename}.')

    def _write_to_original_pdb_file(self, out_df, out_filename):
        hub_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                     f"{self.pdb_name}_original_{out_filename}")

        nodes = self.graph.nodes()

        residue_numbers = [nodes[residue_index][RESIDUE_NUMBER] for residue_index in out_df[ATOM_NUMBER]]
        chain_ids = [nodes[residue_index][CHAIN_ID] for residue_index in out_df[ATOM_NUMBER]]
        insertions = [nodes[residue_index][INSERTION] for residue_index in out_df[ATOM_NUMBER]]
        residue_names = [nodes[residue_index][RESIDUE_NAME] for residue_index in out_df[ATOM_NUMBER]]
        segment_ids = [nodes[residue_index][SEGMENT_ID] for residue_index in out_df[ATOM_NUMBER]]

        centrality_df = out_df.copy()

        centrality_df[RESIDUE_NUMBER] = residue_numbers
        centrality_df[CHAIN_ID] = chain_ids
        centrality_df[INSERTION] = insertions
        centrality_df[RESIDUE_NAME] = residue_names
        centrality_df[SEGMENT_ID] = segment_ids

        centrality_df[KEY] = list(
            zip(centrality_df[RESIDUE_NAME], centrality_df[CHAIN_ID], centrality_df[RESIDUE_NUMBER],
                centrality_df[INSERTION], centrality_df[SEGMENT_ID]))

        residue_to_centrality = centrality_df.set_index(KEY)[CENTRALITY_SCORE].to_dict()

        base_pdb_df = utils.get_base_pdb_df(residue_to=residue_to_centrality,
                                            destination_output_path=self.destination_output_path,
                                            pdb_name=self.pdb_name)

        utils.write_ppdb_to_pdb_file_atom_and_hetatom(ppdb=base_pdb_df, out_filename=hub_file_path)


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

FONT_FAMILY = "Times New Roman"

FONT_STYLES = {
    "title": {"fontsize": 16, "fontfamily": FONT_FAMILY},
    "xlabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "ylabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "zlabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "xtick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "xtick_medium": {"labelsize": 12, "fontfamily": FONT_FAMILY},
    "xtick_small": {"labelsize": 10, "fontfamily": FONT_FAMILY},
    "xtick_small_small": {"labelsize": 8, "fontfamily": FONT_FAMILY},
    "ytick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "tick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "legend": {
        "fontsize": 14,
        "title_fontsize": 14,
        "fontfamily": FONT_FAMILY,
    },
    "colorbar": {
        "label": {"fontsize": 14, "fontfamily": FONT_FAMILY},
        "tick_params": {"labelsize": 14, "fontfamily": FONT_FAMILY},
        "tick_params_label": {"labelsize": 14}
    },
}

EDGE_COLOR = '#1f4e79'
COLOR_PALETTE = ['red', 'gray', 'blue', 'green', 'purple', 'brown', 'teal', 'gold', 'black', 'magenta']



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

RESIDUE_EDGES_FILE: str = "residue_edges.csv"
RESIDUE_AVERAGE_COORDINATES_FILE: str = "residue_average_coordinates.csv"
RESIDUE_AVERAGE_PDB_FILE: str = "residue_average.pdb"
CENTRALITY_CSV_TEMPLATE: str = "centrality_{type}.csv"
CENTRALITY_PDB_TEMPLATE: str = "centrality_{type}.pdb"

HIGH_PERCENTAGE_TEMPLATE: str = "{type}_high_percentage_residues.csv"
FREQUENCY_HIGH_PERCENTAGE_TEMPLATE: str = "frequency_{type}_high_percentage"

TOP_PERCENTAGE_TILE: str = "high_percentage_quantile"
SOURCE_RESIDUE: str = 'source_residue'
TARGET_RESIDUE: str = 'target_residue'
GRAPH_NAME: str = "rinpy"

VALUE: str = "value"

X: str = "x"
Y: str = "y"
Z: str = "z"

REMOVE_HYDROGEN: str = "remove_hydrogen"
IS_CHECKED: str = "is_checked"
CUTOFF: str = "cutoff"

ELEMENT_SYMBOL: str = 'element_symbol'
HYDROGEN_SYMBOL: str = 'H'
PREPROCESS_SAVE_EXT: str = "_pre.pdb"
WITH_HET_ATOM_SAVE_EXT: str = "_with_het_atom.pdb"

RECORD_NAME: str = "record_name"
ATOM_NUMBER: str = "atom_number"
ATOM_NAME: str = "atom_name"
CENTRALITY_SCORE: str = "centrality_score"
B_FACTOR: str = "b_factor"
OCCUPANCY: str = 'occupancy'
RESIDUE_NUMBER: str = "residue_number"
MODE: str = "mode"
RESIDUE_NAME: str = "residue_name"
CHAIN_ID: str = "chain_id"
INSERTION: str = "insertion"
SEGMENT_ID: str = "segment_id"
PDB_ID: str = "pdb_id"
TWO_DECIMAL_FLOAT_FORMATTER: str = '{:.2f}'
THREE_DECIMAL_FLOAT_FORMATTER: str = '{:.3f}'
X_COORD: str = "x_coord"
Y_COORD: str = "y_coord"
Z_COORD: str = "z_coord"
RESIDUE_INDEX: str = "residue_index"
BLANK_4: str = "blank_4"

CA_ATOM_NAME: str = "CA"
P_ATOM_NAME: str = "P"
STAR_PRINT_COUNT: int = 30

PDB_EXT: str = ".pdb"
ENT_EXT: str = ".ent"
EXCEL_EXT: str = ".xlsx"
CSV_EXT: str = ".csv"
TXT_EXT: str = ".txt"
HTML_EXT: str = ".html"
PML_EXT: str = ".pml"
PNG_EXT: str = ".png"
GRAPHML_EXT: str = ".graphml"

DEFAULT_CUTOFF: float = 4.5

ATOM: str = "ATOM"
HETATM: str = "HETATM"
APO: str = "apo"
HOLO: str = "holo"

CALCULATION_OPTION_DEFAULT_JSON_PATH: str = 'calculation_options.json'

PDB_COLUMNS = ['record_name', 'atom_number', 'blank_1', 'atom_name', 'alt_loc',
               'residue_name', 'blank_2', 'chain_id', 'residue_number', 'insertion',
               'blank_3', 'x_coord', 'y_coord', 'z_coord', 'occupancy', 'b_factor',
               'blank_4', 'segment_id', 'element_symbol', 'charge', 'line_idx']

RNA_BASES = {"A", "U", "G", "C"}

DNA_BASES = {"DA", "DT", "DG", "DC"}



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import argparse
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Optional

import matplotlib.colors as mcolors
import networkx as nx
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from scipy.sparse import csr_array
from scipy.sparse.linalg import eigsh

from rinpy import utils, logging_config
from rinpy.communication_path_efficiency import CommunicationPathEfficiency, _RESIDUE_PAIR_TYPE
from rinpy.constants import HIGH_PERCENTAGE_TEMPLATE, ATOM_NUMBER, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID, \
    RESIDUE_NUMBER, INSERTION, SEGMENT_ID, CENTRALITY_CSV_TEMPLATE, X, Y, Z
from rinpy.font_utils import load_fonts_once
from rinpy.log_util import log_time, log_with_stars
from rinpy.style_config import FONT_STYLES, FONT_FAMILY
from rinpy.utils import CentralityType

_NODE_KEY_TYPE = tuple[str, str, int, str, str]  # (residue_name, chain_id, residue_number, insertion, segment_id)
_EDGE_KEY_TYPE = tuple[_NODE_KEY_TYPE, _NODE_KEY_TYPE]
_RESIDUE_KEY_TYPE = tuple[str, int, str, str]  # (chain_id, residue_number, insertion, segment_id)

_RESIDUE_NAME_TARGET = 'residue_name_target'
_RESIDUE_NAME_SOURCE = 'residue_name_source'
_UNCHANGED_LABEL = 'unchanged'
_COUNT = "count"
CENTRALITY_CHANGES_CSV = "centrality_changes.csv"
_VALIDATION_VALUES = [None, '', "''", '""', ""]
_DELTA_MEAN = "delta_mean"
_HTML_3D_SIZE = 18
_WEIGHT = 'weight'
_CHANGE_TYPE = "change_type"
_WEIGHT_CHANGE_PERCENTAGE = "weight_change_percentage"
_WEIGHT_CHANGE = "weight_change"
_TARGET_WEIGHT = "target_weight"
_SOURCE_WEIGHT = "source_weight"
_EDGE_STATUS = "edge_status"
_TARGET_RESIDUE = "target_residue"
_SOURCE_RESIDUE = "source_residue"
_TARGET_NODE_ID = 'target_node_id'
_SOURCE_NODE_ID = 'source_node_id'
_NODE_KEY = 'node_key'
_DELTA = 'delta'
_SCORE = 'score'
_TARGET_SUFFIX = '_target'
_SOURCE_SUFFIX = '_source'
_RESULT_FOLDER_POSTFIX: str = "compare_results"
_CENTRALITY_SCORE_TYPE: str = "{type}_centrality_score"
_CENTRALITY_DELTA: str = "{type}_centrality_delta"

_CENTRALITY_SCORE_COLUMNS = [ATOM_NUMBER, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION,
                             SEGMENT_ID]
_CENTRALITY_SCORE_DTYPES = {ATOM_NUMBER: int, CENTRALITY_SCORE: float, RESIDUE_NAME: str, CHAIN_ID: str,
                            RESIDUE_NUMBER: int,
                            INSERTION: str, SEGMENT_ID: str}
_CENTRALITY_SCORE_SORT_KEYS = [ATOM_NUMBER]
_CENTRALITY_SCORE_COMMON_COLUMNS = [RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]

_ZOOM_CONFIG = {'scrollZoom': True, 'displayModeBar': True}


class ChangeType(Enum):
    INCREASE = "increase"
    DECREASE = "decrease"
    UNCHANGED = "unchanged"
    ADDED = "added"
    REMOVED = "removed"

    def __str__(self):
        return self.value


class EdgeStatusType(Enum):
    COMMON = "common"
    ADDED = "added"
    REMOVED = "removed"
    ALL = "all"

    def __str__(self):
        return self.value


_EDGE_WEIGHT_COLOR_MAP = {
    ChangeType.INCREASE: '#b91c1c',
    ChangeType.DECREASE: '#0761e8',
    ChangeType.UNCHANGED: '#9ca3af',
    ChangeType.ADDED: '#5eead4',
    ChangeType.REMOVED: '#e879f9'
}


class NetworkComparator:
    """ NetworkComparator performs comparative analysis between two protein residue interaction networks (RINs).
        It enables structural comparison and shortest path analysis between specified residue pairs across
        source and target protein structures.

        Parameters
        ----------
        source_input_path : Path
            Directory to the source protein network.

        target_input_path : Path
            Directory to the target protein network.

        output_path : Path
            Directory where analysis results will be saved.

        residue_pairs : list of residue pairs, optional
            List of (source_residue, target_residue) pairs
            used for shortest path analysis.

            Each residue is defined as:
                (chain_id, residue_number, insertion_code, segment_id)

            Format:
                [
                    (('A', 12, '', ''), ('A', 144, '', '')),
                    ...
                ]
        """
    _fonts_initialized = False

    def __init__(self, source_input_path: str,
                 target_input_path: str,
                 output_path: str,
                 residue_pairs: list[_RESIDUE_PAIR_TYPE] | None = None):

        if not NetworkComparator._fonts_initialized:
            load_fonts_once()
            NetworkComparator._fonts_initialized = True

        self.source_input_path = Path(source_input_path)
        self.target_input_path = Path(target_input_path)

        self.source_pdb_name = self.source_input_path.name
        self.source_pdb_name_save = f"{self.source_input_path.name}"
        self.target_pdb_name = self.target_input_path.name
        self.target_pdb_name_save = f"{self.target_input_path.name}"

        self.output_path = Path(
            output_path) / f"{self.source_pdb_name_save}_{self.target_pdb_name_save}_{_RESULT_FOLDER_POSTFIX}"
        self.output_path.mkdir(parents=True, exist_ok=True)

        self.residue_pairs = residue_pairs or []
        logging.info(f"Residue Pairs: {self.residue_pairs}")

        self.source_network = self._get_network(self._find_graphml(self.source_pdb_name, self.source_input_path))
        self.target_network = self._get_network(self._find_graphml(self.target_pdb_name, self.target_input_path))

        self.source_centrality_df = self._get_centrality_score_df(self.source_pdb_name, self.source_input_path)
        self.target_centrality_df = self._get_centrality_score_df(self.target_pdb_name, self.target_input_path)

        # node map: node_id -> key(chain_id, residue_number, insertion, segment_id) # residue_name not included
        self.source_node_ids_to_key_map = self._build_node_id_to_key(self.source_network, False)
        self.target_node_ids_to_key_map = self._build_node_id_to_key(self.target_network, False)

        # Reverse map: key -> node_id
        self.source_node_keys_to_id_map = {v: k for k, v in self.source_node_ids_to_key_map.items()}
        self.target_node_keys_to_id_map = {v: k for k, v in self.target_node_ids_to_key_map.items()}

        self.common_node_keys, self.common_node_ids = self._build_common_residue_nodes(self.source_node_keys_to_id_map,
                                                                                       self.target_node_keys_to_id_map)
        logging_config.clear_logs()

    @log_with_stars("[NetworkComparator] Module")
    @log_time("Network Comparator Analyzer")
    def run(self, num_modes: int = 20) -> None:

        self._process_quantify_changes(num_modes=num_modes)

        self._process_graph_signal_processing()

        if self.residue_pairs:
            self._process_communication_path_efficiency()

        logging.info(f'{"*" * 30} f"{"[NetworkComparator] Module".upper()} ENDS" {"*" * 30}')
        logging_config.save_logs(output_dir=str(self.output_path))

    @log_with_stars("Communication Path Efficiency Analyzer")
    @log_time("Communication Path Efficiency Analyzer")
    def _process_communication_path_efficiency(self) -> None:
        """ Computes communication path efficiency for the given networks"""
        cpe = CommunicationPathEfficiency(source_network=self.source_network,
                                          target_network=self.target_network,
                                          output_path=self.output_path,
                                          residue_pairs=self.residue_pairs,
                                          source_pdb_name=self.source_pdb_name,
                                          target_pdb_name=self.target_pdb_name)
        cpe.compute()

    @log_with_stars("Compute Difference Networks & Quantify Changes")
    @log_time("Compute Difference Networks & Quantify Changes")
    def _process_quantify_changes(self, num_modes: int = 20) -> None:
        """ Computes difference networks between two states and quantify changes in edge weights, centrality measures,
            and community structures.

        Parameters
        ----------
        num_modes : int, default is 20
            specifies the number of non-zero modes used for community structure analysis.
        """
        self._compare_edges()
        self._compare_centrality()
        self._community_structure_analysis(num_modes=num_modes)

    @log_with_stars("Graph Signal Processing")
    @log_time("Graph Signal Processing")
    def _process_graph_signal_processing(self, num_modes: int = 20) -> None:
        """ Performs graph signal processing to analyze how local perturbations propagate
        through the heterogeneous network.

        Parameters
        ----------
        num_modes : int, default is 20
            specifies the number of non-zero modes used for community structure analysis.
        """
        self._compute_graph_signal_processing(num_modes=num_modes)
        self._plot_hub_residues()

    @staticmethod
    def _build_common_residue_nodes(source_keys_to_id: dict[_RESIDUE_KEY_TYPE, int],
                                    target_keys_to_id: dict[_RESIDUE_KEY_TYPE, int]) -> tuple[
        list[_RESIDUE_KEY_TYPE], list[int]]:
        """ builds the common residue node keys and ids. This helps to identify a node either
            using node key or using node id.
        """
        common_keys = set(source_keys_to_id.keys()) & set(target_keys_to_id.keys())
        common_mapping = []
        for key in common_keys:
            common_mapping.append({
                _NODE_KEY: key,  # (chain_id, residue_number, insertion, segment_id)
                _SOURCE_NODE_ID: source_keys_to_id[key],
                _TARGET_NODE_ID: target_keys_to_id[key]
            })
        common_mapping.sort(key=lambda x: (x[_NODE_KEY][0], x[_NODE_KEY][1], x[_NODE_KEY][2], x[_NODE_KEY][3]))

        common_source_ids = [m[_SOURCE_NODE_ID] for m in common_mapping]
        common_target_ids = [m[_TARGET_NODE_ID] for m in common_mapping]

        common_node_keys = [m[_NODE_KEY] for m in common_mapping]
        common_node_ids = list(set(common_source_ids) & set(common_target_ids))

        return common_node_keys, common_node_ids

    @staticmethod
    def _get_network(pdb_network: Path) -> nx.Graph:
        """ load network from the graphml file and return it as a networkx"""
        network = nx.read_graphml(pdb_network)
        node_map = {node: int(node) for node in network.nodes()}
        network = nx.relabel_nodes(network, node_map)
        return network

    @staticmethod
    def _get_high_percentage_score(pdb_name: str, path: Path,
                                   centrality_type: CentralityType = CentralityType.BET) -> pd.DataFrame:
        """ converts the given high percentage file to the dataframe"""
        filename = f"{pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=centrality_type.display_name)}"
        high_centrality_score_path = next(path.glob(filename), None)
        centrality_score_df = utils.get_df(output_file_path=high_centrality_score_path,
                                           columns=_CENTRALITY_SCORE_COLUMNS,
                                           dtypes=_CENTRALITY_SCORE_DTYPES,
                                           sort_keys=_CENTRALITY_SCORE_SORT_KEYS,
                                           sep=";")

        return centrality_score_df

    @staticmethod
    def _get_centrality_df(pdb_name: str, path: Path,
                           centrality_type: CentralityType = CentralityType.BET) -> pd.DataFrame:
        """ converts centrality scores file to the dataframe"""
        df = utils.get_df(output_file_path=next(
            path.glob(f"{pdb_name}_{CENTRALITY_CSV_TEMPLATE.format(type=centrality_type.display_name)}"), None),
            columns=_CENTRALITY_SCORE_COLUMNS,
            dtypes=_CENTRALITY_SCORE_DTYPES,
            sort_keys=_CENTRALITY_SCORE_SORT_KEYS)
        df = df.rename(columns={CENTRALITY_SCORE: _CENTRALITY_SCORE_TYPE.format(type=centrality_type.display_name)})
        return df

    def _get_centrality_score_df(self, pdb_name: str, path: Path) -> pd.DataFrame:
        """ get centrality scores dataframe based on pdb name and the given path directory.
            And also merge all scores into dataframe.
        """
        bet_df = self._get_centrality_df(pdb_name=pdb_name, path=path, centrality_type=CentralityType.BET)
        clos_df = self._get_centrality_df(pdb_name=pdb_name, path=path, centrality_type=CentralityType.CLOS)
        deg_df = self._get_centrality_df(pdb_name=pdb_name, path=path, centrality_type=CentralityType.DEG)

        df = bet_df[[ATOM_NUMBER, _CENTRALITY_SCORE_TYPE.format(
            type=CentralityType.BET.display_name)] + _CENTRALITY_SCORE_COMMON_COLUMNS].merge(
            clos_df[[ATOM_NUMBER, _CENTRALITY_SCORE_TYPE.format(type=CentralityType.CLOS.display_name)]],
            on=ATOM_NUMBER).merge(
            deg_df[[ATOM_NUMBER, _CENTRALITY_SCORE_TYPE.format(type=CentralityType.DEG.display_name)]], on=ATOM_NUMBER)

        df = df[[ATOM_NUMBER, _CENTRALITY_SCORE_TYPE.format(type=CentralityType.BET.display_name),
                 _CENTRALITY_SCORE_TYPE.format(type=CentralityType.CLOS.display_name),
                 _CENTRALITY_SCORE_TYPE.format(type=CentralityType.DEG.display_name),
                 RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]]

        return df

    @staticmethod
    def _find_graphml(pdb_name: str, path: Path) -> Optional[Path]:
        return next(path.glob(f"{pdb_name}_*_network.graphml"), None)

    @staticmethod
    def format_residue_label(key: _NODE_KEY_TYPE) -> str:
        """Format residue key as human-readable label."""
        residue_name, chain_id, residue_number, insertion, segment_id = key
        res_num_insertion = f"{residue_number}{insertion}" if insertion not in _VALIDATION_VALUES else residue_number
        parts = [residue_name, chain_id, str(res_num_insertion)]
        if segment_id and segment_id not in _VALIDATION_VALUES:
            parts.append(segment_id)
        return ", ".join(parts)

    @staticmethod
    def _build_node_id_to_key(network: nx.Graph, include_res_name: bool = True) -> dict:
        """Node ID -> unique residue key tuple mapping."""
        id_to_key = {}
        for node_id, data in network.nodes(data=True):
            residue_name = str(data.get(RESIDUE_NAME))
            chain_id = str(data.get(CHAIN_ID))
            residue_number = int(data.get(RESIDUE_NUMBER))
            insertion = data.get(INSERTION, '')
            if str(insertion) in _VALIDATION_VALUES:
                insertion = ''
            segment_id = data.get(SEGMENT_ID, '')
            if str(segment_id) in _VALIDATION_VALUES:
                segment_id = ''

            if include_res_name:
                id_to_key[node_id] = (residue_name, chain_id, residue_number, str(insertion), str(segment_id))
            else:
                id_to_key[node_id] = (chain_id, residue_number, str(insertion), str(segment_id))

        return id_to_key

    @staticmethod
    def _sort_edge_keys(key_u, key_v):
        """sorts keys based on chain_id, residue_number, insertion, segment_id in order."""

        def sort_key(k):
            residue_name, chain_id, residue_number, insertion, segment_id = k
            return chain_id, residue_number, insertion, segment_id

        return tuple(sorted([key_u, key_v], key=sort_key))

    def _edges_to_residue_keys(self, network: nx.Graph, id_to_key: dict[int, _NODE_KEY_TYPE]) -> set:
        """ it converts sorted residue key tuple sets from network edges. """
        residue_edges = set()
        for u, v in set(network.edges()):
            key_u = id_to_key.get(u)
            key_v = id_to_key.get(v)
            if key_u is None or key_v is None:
                logging.info(f"Warning: node key not found for edge ({u}, {v}), skipping...")
                continue
            residue_edges.add(self._sort_edge_keys(key_u, key_v))
        return residue_edges

    @staticmethod
    def _merge_names(names: set[str]) -> str:
        # deterministic "CYS/GLY"
        return "/".join(sorted(names))

    def _build_merge_map(self, source_edges, target_edges) -> dict[_RESIDUE_KEY_TYPE, str]:
        key_to_names: dict[_RESIDUE_KEY_TYPE, set[str]] = {}

        def _collect(edges) -> None:
            for u, v in edges:
                for node_key in (u, v):
                    key: _RESIDUE_KEY_TYPE = (node_key[1], node_key[2], node_key[3], node_key[4])
                    key_to_names.setdefault(key, set()).add(node_key[0])  # residue_name

        _collect(source_edges)
        _collect(target_edges)

        return {k: self._merge_names(v) for k, v in key_to_names.items()}

    @staticmethod
    def _rewrite_edges(edges, merge_map: dict[_RESIDUE_KEY_TYPE, str]) -> set[_EDGE_KEY_TYPE]:

        out: set[_EDGE_KEY_TYPE] = set()

        for u, v in edges:
            def rewrite(n: _NODE_KEY_TYPE) -> _NODE_KEY_TYPE:
                residue_name, chain_id, residue_number, insertion, segment_id = n
                merged_residue_name = merge_map.get((chain_id, residue_number, insertion, segment_id), residue_name)
                return merged_residue_name, chain_id, residue_number, insertion, segment_id

            uu = rewrite(u)
            vv = rewrite(v)
            out.add((uu, vv))

        return out

    @staticmethod
    def _format_common_edge_residue(node_key: _NODE_KEY_TYPE) -> str:
        residue_name, chain_id, residue_number, insertion, segment_id = node_key
        res_part = f"{residue_number}{insertion}" if insertion else f"{residue_number}"
        base = f"{residue_name}, {chain_id}, {res_part}"
        if segment_id:
            return f"{base},{segment_id}"
        return base

    @log_time("Compare Edges")
    def _compare_edges(self) -> pd.DataFrame:
        """ compares the edges between source and target networks.
            It plots the distribution in 2D and 3D format.
        """
        source_id_to_key_with_res_name = self._build_node_id_to_key(self.source_network)
        target_id_to_key_with_res_name = self._build_node_id_to_key(self.target_network)

        source_id_to_key = self._build_node_id_to_key(self.source_network, False)
        target_id_to_key = self._build_node_id_to_key(self.target_network, False)

        # Reverse map: key -> node_id
        source_key_to_id = {v: k for k, v in source_id_to_key.items()}
        target_key_to_id = {v: k for k, v in target_id_to_key.items()}

        source_edges_keys = self._edges_to_residue_keys(self.source_network, source_id_to_key_with_res_name)
        target_edges_keys = self._edges_to_residue_keys(self.target_network, target_id_to_key_with_res_name)

        merge_map = self._build_merge_map(source_edges_keys, target_edges_keys)

        source_edge_pairs = self._rewrite_edges(source_edges_keys, merge_map)
        target_edge_pairs = self._rewrite_edges(target_edges_keys, merge_map)

        common_edges = source_edge_pairs & target_edge_pairs
        added_edges = target_edge_pairs - source_edge_pairs
        removed_edges = source_edge_pairs - target_edge_pairs

        rows = []

        for key_u, key_v in common_edges:
            u_key = key_u[1:]
            v_key = key_v[1:]

            src_u_id = source_key_to_id[u_key]
            src_v_id = source_key_to_id[v_key]

            tgt_u_id = target_key_to_id[u_key]
            tgt_v_id = target_key_to_id[v_key]

            source_weight = self.source_network[src_u_id][src_v_id].get(_WEIGHT, 1.0)
            target_weight = self.target_network[tgt_u_id][tgt_v_id].get(_WEIGHT, 1.0)

            weight_change = target_weight - source_weight

            weight_change_percentage = (
                    (target_weight - source_weight) / source_weight * 100) if source_weight != 0 else 0

            if weight_change > 0:
                change_type = ChangeType.INCREASE
            elif weight_change < 0:
                change_type = ChangeType.DECREASE
            else:
                change_type = ChangeType.UNCHANGED

            rows.append([
                self._format_common_edge_residue(key_u),
                self._format_common_edge_residue(key_v),
                EdgeStatusType.COMMON,
                source_weight,
                target_weight,
                weight_change,
                weight_change_percentage,
                change_type
            ])

        for key_u, key_v in added_edges:
            tgt_u_id = target_key_to_id[key_u[1:]]
            tgt_v_id = target_key_to_id[key_v[1:]]
            target_weight = self.target_network[tgt_u_id][tgt_v_id].get(_WEIGHT, 1.0)

            rows.append([
                self._format_common_edge_residue(key_u),
                self._format_common_edge_residue(key_v),
                EdgeStatusType.ADDED,
                0.0,
                target_weight,
                target_weight,
                float('inf'),
                ChangeType.ADDED
            ])

        for key_u, key_v in removed_edges:
            src_u_id = source_key_to_id[key_u[1:]]
            src_v_id = source_key_to_id[key_v[1:]]
            source_weight = self.source_network[src_u_id][src_v_id].get(_WEIGHT, 1.0)

            rows.append([
                self._format_common_edge_residue(key_u),
                self._format_common_edge_residue(key_v),
                EdgeStatusType.REMOVED,
                source_weight,
                0.0,
                -source_weight,
                float('-inf'),
                ChangeType.REMOVED
            ])

        _EDGE_WEIGHT_COLUMNS = [_SOURCE_RESIDUE, _TARGET_RESIDUE, _EDGE_STATUS, _SOURCE_WEIGHT, _TARGET_WEIGHT,
                                _WEIGHT_CHANGE, _WEIGHT_CHANGE_PERCENTAGE, _CHANGE_TYPE]

        df = pd.DataFrame(rows, columns=_EDGE_WEIGHT_COLUMNS)

        logging.info("Creating edge weight distribution 2D plot...")
        self._visualize_edge_weights_distribution_2d(df=df)

        logging.info("Creating edge weight distribution 3D plot...")
        self._visualize_edge_weights_distribution_3d(df=df)

        return df

    def _visualize_edge_weights_distribution_2d(self, df: pd.DataFrame = None) -> None:
        """ Visualize edge weight distributions and changes """

        if df is None:
            logging.info("Given dataframe is None.")
            return

        df_processed = df.copy()

        if df_processed.empty:
            logging.info("No edges found.")
            return

        plt.figure(figsize=(12, 8))

        for change_type, color in _EDGE_WEIGHT_COLOR_MAP.items():
            mask = df_processed[_CHANGE_TYPE] == change_type
            plt.scatter(df_processed.loc[mask, _SOURCE_WEIGHT],
                        df_processed.loc[mask, _TARGET_WEIGHT],
                        alpha=0.55, s=22, c=color, label=change_type.value.capitalize(),
                        edgecolors='none')

        max_val = max(df_processed[_SOURCE_WEIGHT].max(), df_processed[_TARGET_WEIGHT].max())
        plt.plot([0, max_val], [0, max_val], '--', color='#4b5563', linewidth=1.5, label='No change')

        padding = max_val * 0.02
        plt.xlim(0 - padding, max_val + padding)
        plt.ylim(0 - padding, max_val + padding)
        plt.gca().set_aspect('equal', adjustable='box')

        plt.xlabel(f'Edge weight ({self.source_pdb_name})', fontsize=FONT_STYLES['xlabel']['fontsize'])
        plt.ylabel(f'Edge weight ({self.target_pdb_name})', fontsize=FONT_STYLES['ylabel']['fontsize'])
        plt.title(f'{self.source_pdb_name} vs {self.target_pdb_name} Edge Weights',
                  fontsize=FONT_STYLES['title']['fontsize'])
        plt.legend(fontsize=FONT_STYLES['legend']['fontsize'])
        plt.grid(alpha=0.3)
        plt.tight_layout()

        output_file = self.output_path / 'edge_weight_distributions_2d.png'
        plt.savefig(output_file, dpi=300, bbox_inches='tight')
        plt.close()
        logging.info(f"Saved: {output_file}")

    def _visualize_edge_weights_distribution_3d(self, df=None) -> None:
        """ Interactive Plotly visualization of edge weight changes in a 3D format. """
        if df is None:
            logging.info("Given dataframe is None.")
            return
        df_processed = df.copy()

        if df_processed.empty:
            logging.info("No common edges found.")
            return

        fig = go.Figure()

        for change_type, color in _EDGE_WEIGHT_COLOR_MAP.items():
            subset = df_processed[df_processed[_CHANGE_TYPE] == change_type]
            if subset.empty:
                continue

            fig.add_trace(go.Scatter(
                x=subset[_SOURCE_WEIGHT],
                y=subset[_TARGET_WEIGHT],
                mode='markers',
                marker=dict(size=7, color=color, opacity=0.7),
                name=change_type.value.capitalize(),
                text=[
                    f"{self._format_centrality_node_label(row[_SOURCE_RESIDUE])} ↔ "
                    f"{self._format_centrality_node_label(row[_TARGET_RESIDUE])}<br>"
                    f"{self.source_pdb_name}: {row[_SOURCE_WEIGHT]:.3f}<br>"
                    f"{self.target_pdb_name}: {row[_TARGET_WEIGHT]:.3f}<br>"
                    f"Δ: {row[_WEIGHT_CHANGE]:.3f}"
                    for _, row in subset.iterrows()],
                hovertemplate='%{text}<extra></extra>'
            ))

        # Diagonal line
        max_val = max(df_processed[_SOURCE_WEIGHT].max(), df_processed[_TARGET_WEIGHT].max())
        fig.add_trace(go.Scatter(
            x=[0, max_val], y=[0, max_val],
            mode='lines',
            line=dict(color='#4b5563', dash='dash', width=1.5),
            name='No change',
            showlegend=False
        ))

        fig.update_layout(
            title=dict(text=f'{self.source_pdb_name} vs {self.target_pdb_name} Edge Weights', x=0.5, xanchor='center'),
            xaxis_title=f'Edge weight ({self.source_pdb_name})',
            yaxis_title=f'Edge weight ({self.target_pdb_name})',
            height=900,
            width=900,
            showlegend=True,
            legend=dict(
                x=1.12,
                y=0.90,
                xanchor="right",
                yanchor="top"
            ),
            xaxis=dict(
                domain=[0.08, 0.88]
            ),
            yaxis=dict(
                domain=[0.08, 0.92]
            ),
            template="simple_white",
            font=dict(size=18, family=FONT_FAMILY, color="black"),
            margin=dict(l=120, r=120, t=100, b=120),
            hovermode="closest"
        )

        output_file = self.output_path / 'edge_weight_distributions_3d.html'
        fig.write_html(output_file, config=_ZOOM_CONFIG)
        logging.info(f"Saved: {output_file}")

    @staticmethod
    def _format_centrality_node_label(node: str, splitter: str = ",") -> str:
        parts = [p.strip() for p in str(node).split(splitter)]
        if len(parts) < 3:
            return str(node).strip()
        residue_name = "/".join(part.capitalize() for part in parts[0].split("/"))
        chain_id = parts[1]
        residue_number = parts[2]
        segment_id = parts[3] if len(parts) >= 4 and parts[3] else None
        label = f"{chain_id}: {residue_name}{residue_number}"
        if segment_id:
            label += f" ({segment_id})"
        return label

    @staticmethod
    def _get_contrast_color(rgba):
        r, g, b = rgba[:3]
        luminance = 0.299 * r + 0.587 * g + 0.114 * b
        return "black" if luminance > 0.6 else "white"

    def _plot_centrality_score_comparison_3d(self, centrality_type=CentralityType.BET, output_name=None,
                                             df=None) -> None:

        if output_name is None:
            output_name = f"centrality_{centrality_type.display_name}_comparison_3d.html"

        source_column = f"source_{centrality_type.display_name}"
        target_column = f"target_{centrality_type.display_name}"
        delta_column = f"delta_{centrality_type.display_name}"

        gx, gy = source_column, target_column

        df_processed = df.copy()

        grouped_df = (
            df_processed.groupby([gx, gy], dropna=False)
            .agg(
                nodes=("node", lambda s: list(s)),
                count=("node", "size"),
                delta_mean=(delta_column, "mean")
            )
            .reset_index()
        )

        grouped_df["nodes_html"] = grouped_df["nodes"].apply(lambda xs: "<br>".join(x for x in xs))

        fig = go.Figure()

        cmap = plt.cm.RdBu_r
        norm = mcolors.Normalize(vmin=grouped_df[_DELTA_MEAN].min(),
                                 vmax=grouped_df[_DELTA_MEAN].max())

        rgba_colors = cmap(norm(grouped_df[_DELTA_MEAN].values))

        text_colors = [self._get_contrast_color(rgba) for rgba in rgba_colors]

        max_val = max(df_processed[source_column].max(), df_processed[target_column].max())
        fig.add_trace(go.Scatter(
            x=[0, max_val],
            y=[0, max_val],
            mode="lines",
            line=dict(dash="dash", color="gray", width=1.5),
            name="No Change"
        ))

        fig.add_trace(go.Scatter(
            x=grouped_df[gx],
            y=grouped_df[gy],
            mode="markers+text",
            text=np.where(grouped_df[_COUNT] > 1, grouped_df[_COUNT].astype(str), ""),
            textposition="middle center",
            textfont=dict(size=12, color=text_colors),
            marker=dict(
                size=np.clip(10 + 3 * grouped_df[_COUNT], 10, 34),
                color=grouped_df[_DELTA_MEAN],
                colorscale="RdBu_r",
                line=dict(width=0.7, color="black"),
                colorbar=dict(
                    title="",
                    thickness=20,
                    len=0.75,
                    x=1.02,
                    ticks="outside",
                    tickwidth=2,
                    ticklen=6,
                    tickcolor="gray",
                    tickfont=dict(size=16, color="black", family=FONT_FAMILY)
                ),
                opacity=0.8
            ),
            hovertemplate="<b>Residues</b><br>%{customdata[0]}<br><br>" +
                          f"{self.source_pdb_name}: %{{x:.6f}}<br>" +
                          f"{self.target_pdb_name}: %{{y:.6f}}<br>" +
                          "count: %{customdata[1]}<br>" +
                          f"Δ: %{{marker.color:.4f}}" +
                          "<extra></extra>",
            customdata=np.stack([
                grouped_df["nodes_html"],
                grouped_df[_COUNT],
                grouped_df[_DELTA_MEAN]
            ], axis=1),
            name="Residues"
        ))

        fig.add_annotation(
            x=1.16,
            y=0.5,
            xref="paper",
            yref="paper",
            text=f"Δ {centrality_type.display_name.capitalize()}",
            textangle=90,
            showarrow=False,
            font=dict(size=20, family=FONT_FAMILY, color="black"),
            align="center"
        )

        fig.update_layout(
            width=1000,
            height=900,
            title=dict(
                text=f"{centrality_type.display_name.capitalize()} Centrality Comparison",
                x=0.5,
                xanchor="center"
            ),
            xaxis=dict(
                title=f"{centrality_type.display_name.capitalize()} ({self.source_pdb_name})",
                domain=[0.08, 0.88]
            ),
            yaxis=dict(
                title=f"{centrality_type.display_name.capitalize()} ({self.target_pdb_name})",
                domain=[0.08, 0.92]
            ),
            template="simple_white",
            font=dict(size=17, family=FONT_FAMILY),
            margin=dict(l=120, r=120, t=100, b=120),
            hovermode="closest"
        )

        fig.write_html(os.path.join(self.output_path, output_name), config=_ZOOM_CONFIG)

    def _plot_centrality_score_comparison_2d(self, centrality_type=CentralityType.BET, output_name=None,
                                             df=None) -> None:

        if output_name is None:
            output_name = f"centrality_{centrality_type.display_name}_comparison_2d.png"

        source_column = f"source_{centrality_type.display_name}"
        target_column = f"target_{centrality_type.display_name}"
        delta_column = f"delta_{centrality_type.display_name}"

        df_processed = df.copy()
        s_g = f"_source_{centrality_type.display_name}"
        t_g = f"_target_{centrality_type.display_name}"
        df_processed[s_g] = df_processed[source_column]
        df_processed[t_g] = df_processed[target_column]

        grouped_df = (
            df_processed.groupby([s_g, t_g], dropna=False)
            .agg(
                delta_mean=(delta_column, "mean"),
                count=("node", "size"),
                nodes=("node", lambda s: list(s))
            )
            .reset_index()
        )

        fig, ax = plt.subplots(figsize=(8, 8))

        v_min, v_max = grouped_df[_DELTA_MEAN].min(), grouped_df[_DELTA_MEAN].max()
        norm = plt.Normalize(vmin=v_min, vmax=v_max)
        cmap = plt.cm.RdBu_r

        rgba_colors = cmap(norm(grouped_df[_DELTA_MEAN].values))
        text_colors = [self._get_contrast_color(rgba) for rgba in rgba_colors]
        sizes = np.clip(30 + 15 * grouped_df[_COUNT], 40, 260)
        ax.scatter(
            grouped_df[s_g], grouped_df[t_g],
            c=grouped_df[_DELTA_MEAN],
            s=sizes,
            cmap=cmap,
            norm=norm,
            alpha=0.75,
            edgecolors="black",
            linewidth=0.6,
            zorder=2
        )

        for x, y, cnt, text_color in zip(grouped_df[s_g], grouped_df[t_g], grouped_df[_COUNT], text_colors):
            if cnt > 1:
                ax.text(x, y, str(cnt), ha="center", va="center", fontsize=8, color=text_color, zorder=10)

        max_val = max(df_processed[source_column].max(), df_processed[target_column].max())
        ax.plot([0, max_val], [0, max_val], 'k--', linewidth=1, alpha=0.5, label="No Change")

        ax.set_xlabel(f'{centrality_type.display_name.capitalize()} Centrality ({self.source_pdb_name})',
                      fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel(f'{centrality_type.display_name.capitalize()} Centrality ({self.target_pdb_name})',
                      fontdict=FONT_STYLES["ylabel"])
        ax.tick_params(axis='both', labelsize=FONT_STYLES['xtick_medium']['labelsize'])
        ax.set_title(f'{centrality_type.display_name.capitalize()} Centrality Comparison',
                     fontdict=FONT_STYLES["title"])
        ax.legend(loc='upper left')
        ax.grid(True, alpha=0.3)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        cbar = plt.colorbar(sm, ax=ax)
        cbar.set_label(f'Δ {centrality_type.display_name.capitalize()} (mean)', rotation=270, labelpad=20,
                       fontsize=12)
        cbar.ax.tick_params(labelsize=12)

        plt.tight_layout()
        plt.savefig(os.path.join(self.output_path, output_name), dpi=300, bbox_inches='tight')
        plt.close()

    def _plot_delta_histogram(self, data: np.ndarray, x_label: str, filename: str):
        """Plot histogram distribution."""
        plt.figure()
        plt.hist(data, bins=40, edgecolor='black', alpha=0.7)
        plt.xlabel(x_label)
        plt.ylabel("Frequency")
        plt.tight_layout()
        plt.savefig(os.path.join(self.output_path, filename), dpi=300, bbox_inches='tight')
        plt.close()

    @log_time("Compare Centrality Scores")
    def _compare_centrality(self):
        merge_keys = [CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]
        score_columns = [_CENTRALITY_SCORE_TYPE.format(type=CentralityType.BET.display_name),
                         _CENTRALITY_SCORE_TYPE.format(type=CentralityType.CLOS.display_name),
                         _CENTRALITY_SCORE_TYPE.format(type=CentralityType.DEG.display_name)]

        df = self.source_centrality_df.merge(self.target_centrality_df, on=merge_keys,
                                             suffixes=(_SOURCE_SUFFIX, _TARGET_SUFFIX))

        for column in score_columns:
            df[column.replace(_SCORE, _DELTA)] = df[f'{column}{_TARGET_SUFFIX}'] - df[f'{column}{_SOURCE_SUFFIX}']

        header_columns = ["node", "source_betweenness", "target_betweenness", "delta_betweenness", "source_closeness",
                          "target_closeness", "delta_closeness", "source_degree", "target_degree", "delta_degree"]

        rows = []

        for _, row in df.iterrows():
            res_name = (
                f"{row[_RESIDUE_NAME_SOURCE]}/{row[_RESIDUE_NAME_TARGET]}"
                if row[_RESIDUE_NAME_SOURCE] != row[_RESIDUE_NAME_TARGET]
                else row[_RESIDUE_NAME_SOURCE]
            )
            node_label = self._format_centrality_node_label(self.format_residue_label(
                (
                    res_name,
                    row[CHAIN_ID],
                    row[RESIDUE_NUMBER],
                    "" if pd.isna(row[INSERTION]) or row[INSERTION] in _VALIDATION_VALUES else str(row[INSERTION]),
                    "" if pd.isna(row[SEGMENT_ID]) or row[SEGMENT_ID] in _VALIDATION_VALUES else str(row[SEGMENT_ID]),
                )
            ))

            rows.append([
                node_label,
                row["betweenness_centrality_score_source"],
                row["betweenness_centrality_score_target"],
                row["betweenness_centrality_delta"],
                row["closeness_centrality_score_source"],
                row["closeness_centrality_score_target"],
                row["closeness_centrality_delta"],
                row["degree_centrality_score_source"],
                row["degree_centrality_score_target"],
                row["degree_centrality_delta"]
            ])

        delta_columns = {
            _CENTRALITY_DELTA.format(type=CentralityType.BET.display_name): 'Δ Betweenness Centrality',
            _CENTRALITY_DELTA.format(type=CentralityType.CLOS.display_name): 'Δ Closeness Centrality',
            _CENTRALITY_DELTA.format(type=CentralityType.DEG.display_name): 'Δ Degree Centrality'
        }

        for col, x_label in delta_columns.items():
            self._plot_delta_histogram(df[col].values, x_label, f"{col}_distribution.png")

        saved_df = pd.DataFrame(rows, columns=header_columns)
        saved_df.to_csv(self.output_path / CENTRALITY_CHANGES_CSV, index=False, header=True)
        for centrality_type in [CentralityType.BET, CentralityType.CLOS, CentralityType.DEG]:
            self._plot_centrality_score_comparison_2d(centrality_type=centrality_type, df=saved_df)
            self._plot_centrality_score_comparison_3d(centrality_type=centrality_type, df=saved_df)

    def _get_node_ids(self, network: nx.Graph) -> list[int]:
        node_ids: list[int] = []
        for node_key in self.common_node_keys:
            chain_id, residue_number, insertion, segment_id = node_key
            for node_id, attrs in network.nodes(data=True):
                if (attrs.get(CHAIN_ID) == chain_id and int(attrs.get(RESIDUE_NUMBER)) == residue_number
                        and (attrs.get(INSERTION) or "") == insertion and (attrs.get(SEGMENT_ID) or "") == segment_id):
                    node_ids.append(node_id)
                    break
        return sorted(node_ids)

    def _compute_modes(self, network: nx.Graph, num_modes: int) -> tuple[np.ndarray, np.ndarray, csr_array] | None:

        nodes = self._get_node_ids(network)

        laplacian = nx.laplacian_matrix(network, nodelist=nodes).astype(float)

        if not nx.is_connected(network):
            logging.error("Network must be connected for spectral partitioning.")
            return

        k = num_modes + 1
        eigen_values, eigen_vectors = eigsh(laplacian, k=k, which="SM")

        idx = np.argsort(eigen_values)
        eigen_values = eigen_values[idx]
        eigen_vectors = eigen_vectors[:, idx]

        spectral_modes = eigen_vectors[:, 1:k]  # Fiedler +  {num_modes-1} more num_modes included

        return spectral_modes, eigen_values[1:k], laplacian

    def _save_eigen_vectors_and_values(self, spectral_modes: np.ndarray, eigen_values: np.ndarray = None,
                                       prefix: str = "source"):
        num_modes = spectral_modes.shape[1]
        mode_columns = [f"Mode_{i + 1}" for i in range(num_modes)]
        df_modes = pd.DataFrame(spectral_modes, columns=mode_columns)
        df_modes.to_csv(self.output_path / f"{prefix}_eigenvector.csv", index=False)
        if eigen_values is not None:
            df_eigenvalues = pd.DataFrame({
                "Mode": [f"Mode_{i + 1}" for i in range(len(eigen_values))],
                "Eigenvalue": eigen_values
            })

            df_eigenvalues.to_csv(self.output_path / f"{prefix}_eigenvalues.csv", index=False)

    @log_time("Community Structure Analysis")
    def _community_structure_analysis(self, num_modes=10, save_plot=True):
        source_modes, source_eigen_values, _ = self._compute_modes(self.source_network, num_modes=num_modes)
        target_modes, target_eigen_values, _ = self._compute_modes(self.target_network, num_modes=num_modes)

        self._save_eigen_vectors_and_values(source_modes, source_eigen_values, prefix=self.source_pdb_name)
        self._save_eigen_vectors_and_values(target_modes, target_eigen_values, prefix=self.target_pdb_name)

        target_modes_norm = np.where(target_modes < 0, -1, 1)
        source_modes_norm = np.where(source_modes < 0, -1, 1)
        self._save_eigen_vectors_and_values(target_modes_norm, None, prefix=f"{self.source_pdb_name}_norm")
        self._save_eigen_vectors_and_values(source_modes_norm, None, prefix=f"{self.target_pdb_name}_norm")

        num_nodes = target_modes_norm.shape[0]
        overlap = np.abs(target_modes_norm.T @ source_modes_norm) / num_nodes

        if save_plot:
            self._save_community_structure_analysis(overlap)

    def _save_community_structure_analysis(self, overlap: np.ndarray) -> None:
        fig, ax = plt.subplots(figsize=(8, 7))
        im = ax.imshow(overlap, vmin=0, vmax=1, cmap='jet', origin='lower')
        ticks = np.arange(overlap.shape[0])
        labels = np.arange(1, overlap.shape[0] + 1)
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, fontdict=FONT_STYLES["xlabel"])
        ax.set_yticks(ticks)
        ax.set_yticklabels(labels, fontdict=FONT_STYLES["xlabel"])
        ax.set_xlabel(f'Modes ({self.source_pdb_name})', fontsize=18)
        ax.set_ylabel(f'Modes ({self.target_pdb_name})', fontsize=18)
        ax.set_title('Community Structure', fontsize=22, pad=12)
        cbar = fig.colorbar(im, ax=ax, shrink=0.96, aspect=25, pad=0.03)
        cbar.ax.tick_params(**FONT_STYLES["colorbar"]["tick_params_label"])
        plt.tight_layout()
        plt.savefig(self.output_path / 'community_structure.png', dpi=300, bbox_inches='tight')
        plt.close()

    @staticmethod
    def _parse_label(label: str) -> tuple[str, str, str, Optional[str]]:
        label = label.strip().replace("'", "").replace('"', "")
        parts = label.split(":")
        residue_name = parts[0]
        chain_id = parts[1]
        residue_number = parts[2]
        segment_id = parts[3] if len(parts) == 4 else None
        return residue_name, chain_id, residue_number, segment_id

    @log_time("Computing Graph Signal Processing")
    def _compute_graph_signal_processing(self, num_modes=10, save_plot: bool = True, has_title: bool = False):
        source_modes, source_eigen_values, _ = self._compute_modes(network=self.source_network, num_modes=num_modes)
        target_modes, target_eigen_values, _ = self._compute_modes(network=self.target_network, num_modes=num_modes)

        shift_values = ((target_eigen_values - source_eigen_values) / source_eigen_values) * 100

        if save_plot:
            self._save_shift_signals(has_title, shift_values)

    def _save_shift_signals(self, has_title: bool, shift_values: list[float], has_marker_title: bool = False) -> None:
        modes = np.arange(1, len(shift_values) + 1)
        shift_values = np.array(shift_values, dtype=float)

        plt.figure()
        plt.plot(modes, shift_values, marker='o')
        plt.axhline(0, color='black', linewidth=0.8, alpha=0.3, linestyle='--')

        plt.xlabel("Modes", fontsize=18)
        plt.ylabel("Eigenvalue shift (%)", fontsize=18)

        if has_title:
            plt.title("Percent Eigenvalue Shift per Mode", fontsize=22)

        plt.xticks(modes)

        if np.any(shift_values > 0) and np.any(shift_values < 0):
            y_min = min(shift_values)
            y_max = max(shift_values)
            margin = (y_max - y_min) * 0.1
            plt.ylim(y_min - margin, y_max + margin)

        if has_marker_title:
            for x, y in zip(modes, shift_values):
                offset = 1.2 if y < 0 else 0.8
                plt.text(x, y + offset, f"{y:.2f}", ha='center', va='bottom' if y >= 0 else 'top', fontsize=9)

        plt.tight_layout()
        plt.grid(alpha=0.4)
        plt.savefig(self.output_path / 'shift_modes.png', dpi=300, bbox_inches='tight')
        plt.close()

    @log_time("Plotting Hub Residues")
    def _plot_hub_residues(self):
        source_high_bet_df = self._get_high_percentage_score(pdb_name=self.source_pdb_name,
                                                             path=self.source_input_path,
                                                             centrality_type=CentralityType.BET)

        target_high_bet_df = self._get_high_percentage_score(pdb_name=self.target_pdb_name,
                                                             path=self.target_input_path,
                                                             centrality_type=CentralityType.BET)

        self._visualize_high_betweenness_2d(pdb_name=self.source_pdb_name,
                                            pdb_save_name=self.source_pdb_name_save,
                                            network=self.source_network,
                                            high_betweenness_df=source_high_bet_df,
                                            save=True,
                                            node_keys_to_id_map=self.source_node_keys_to_id_map)
        self._visualize_high_betweenness_2d(pdb_name=self.target_pdb_name,
                                            pdb_save_name=self.target_pdb_name_save,
                                            network=self.target_network,
                                            high_betweenness_df=target_high_bet_df,
                                            save=True,
                                            node_keys_to_id_map=self.target_node_keys_to_id_map)

        self._visualize_high_betweenness_3d(pdb_name=self.source_pdb_name,
                                            pdb_save_name=self.source_pdb_name_save,
                                            network=self.source_network,
                                            high_betweenness_df=source_high_bet_df,
                                            save=True,
                                            node_keys_to_id_map=self.source_node_keys_to_id_map)
        self._visualize_high_betweenness_3d(pdb_name=self.target_pdb_name,
                                            pdb_save_name=self.target_pdb_name_save,
                                            network=self.target_network,
                                            high_betweenness_df=target_high_bet_df,
                                            save=True,
                                            node_keys_to_id_map=self.target_node_keys_to_id_map)

    @staticmethod
    def _format_node_label_by_node_id(network: nx.Graph, node: int) -> str:
        """ Formats node label like residue_name:chain_id:residue_number+insertion:segment_id """
        data = network.nodes[node]
        residue_name = data.get(RESIDUE_NAME, '')
        chain_id = data.get(CHAIN_ID, '')
        residue_number = str(data.get(RESIDUE_NUMBER, ''))

        insertion = data.get(INSERTION, '')
        if insertion and insertion not in _VALIDATION_VALUES:
            residue_number += str(insertion)

        segment_id = data.get(SEGMENT_ID, '')
        if segment_id and segment_id not in _VALIDATION_VALUES:
            label = f"{residue_name}:{chain_id}:{residue_number}:{segment_id}"
        else:
            label = f"{residue_name}:{chain_id}:{residue_number}"

        return label

    @staticmethod
    def _generate_node_key(row: pd.Series) -> _RESIDUE_KEY_TYPE:
        """ generate node key from attributes of a DataFrame row."""

        def _norm(v) -> str:
            v = '' if str(v) in [str(x) for x in _VALIDATION_VALUES] else v
            return '' if v is None else str(v)

        return row[CHAIN_ID], row[RESIDUE_NUMBER], _norm(row.get(INSERTION, '')), _norm(row.get(SEGMENT_ID, ''))

    def _find_high_betweenness_residues(self, high_betweenness_df: pd.DataFrame,
                                        node_keys_to_id_map: dict[_RESIDUE_KEY_TYPE, int]):
        highlight_nodes = set()
        scores = {}
        for _, row in high_betweenness_df.iterrows():
            node_id = node_keys_to_id_map.get(self._generate_node_key(row), None)
            if node_id is not None:
                highlight_nodes.add(node_id)
                scores[node_id] = row[CENTRALITY_SCORE]
        return highlight_nodes, scores

    def _visualize_high_betweenness_2d(self, pdb_name: str, pdb_save_name: str, network: nx.Graph,
                                       high_betweenness_df: pd.DataFrame, save: bool = True,
                                       node_keys_to_id_map: dict[_RESIDUE_KEY_TYPE, int] = None) -> None:
        """
        2D network visualization — high betweenness residues larger and colored while other residue nodes are smaller
        and colored with grey.

        Parameters
        ----------
        pdb_name : str
        pdb filename : str
        network : networkx.Graph
        high_betweenness_df : pd.DataFrame
            columns: atom_number, centrality_score, residue_name, chain_id, residue_number, insertion, segment_id
        save : bool
            save the figure if it is True.
        """

        highlight_residues, scores = self._find_high_betweenness_residues(high_betweenness_df, node_keys_to_id_map)

        pos = {}
        for node, data in network.nodes(data=True):
            pos[node] = (data.get(X, 0), data.get(Y, 0))

        all_residue_nodes = list(network.nodes())
        node_colors = []
        node_sizes = []
        node_labels = {}

        for node in all_residue_nodes:
            if node in highlight_residues:
                node_colors.append('#0761E8')
                node_sizes.append(120)
                node_labels[node] = self._format_centrality_node_label(
                    self._format_node_label_by_node_id(network, node),
                    splitter=":")
            else:
                node_colors.append('#d1d5db')
                node_sizes.append(15)

        edge_colors = []
        edge_widths = []
        for u, v in network.edges():
            if u in highlight_residues and v in highlight_residues:
                edge_colors.append('black')
                edge_widths.append(2.0)
            else:
                edge_colors.append('#9ca3af')
                edge_widths.append(0.5)

        plt.figure(figsize=(14, 10))

        nx.draw_networkx_edges(network, pos,
                               edgelist=network.edges(),
                               edge_color=edge_colors,
                               width=edge_widths,
                               alpha=0.4)

        normal_residue_nodes = [n for n in all_residue_nodes if n not in highlight_residues]
        nx.draw_networkx_nodes(network, pos,
                               nodelist=normal_residue_nodes,
                               node_color='#4b5563',
                               node_size=30,
                               alpha=0.7,
                               edgecolors='#374151',
                               linewidths=0.5)

        highlight_list = list(highlight_residues)
        h_scores = [scores.get(n, 0) for n in highlight_list]

        min_s, max_s = min(h_scores), max(h_scores)

        cmap = plt.matplotlib.colors.LinearSegmentedColormap.from_list(CentralityType.BET.display_name,
                                                                       ['#0761E8', '#b91c1c'])
        norm = plt.Normalize(vmin=min_s, vmax=max_s)
        h_colors = [cmap(norm(s)) for s in h_scores]

        nx.draw_networkx_nodes(network, pos,
                               nodelist=highlight_list,
                               node_color=h_colors,
                               node_size=90,
                               edgecolors='#374151',
                               linewidths=1,
                               alpha=0.8)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        cbar = plt.colorbar(sm, ax=plt.gca(), shrink=0.6, pad=0.02)
        cbar.set_label('Betweenness Centrality', fontsize=12, labelpad=20, rotation=270)

        label_pos = {node: (x, y + (max(pos.values(), key=lambda p: p[1])[1] -
                                    min(pos.values(), key=lambda p: p[1])[1]) * 0.02)
                     for node, (x, y) in pos.items() if node in node_labels}

        nx.draw_networkx_labels(network, label_pos,
                                labels=node_labels,
                                font_size=9,
                                font_color='#111827',
                                font_family=FONT_FAMILY
                                )

        plt.title(f'{pdb_name} — High Betweenness Centrality Residues', fontsize=18)
        plt.axis('off')
        plt.margins(0.02)

        if save:
            output_file = self.output_path / f'{pdb_save_name}_high_betweenness_network_2d.png'
            plt.savefig(output_file, dpi=300, bbox_inches='tight', pad_inches=0.1)
            logging.info(f"{output_file} has been saved!")

        plt.close()

    def _visualize_high_betweenness_3d(self, pdb_name: str, pdb_save_name: str, network: nx.Graph,
                                       high_betweenness_df: pd.DataFrame, save: bool = True,
                                       node_keys_to_id_map: dict[_RESIDUE_KEY_TYPE, int] = None) -> None:
        """
        3D interactive Plotly visualization — high betweenness residues colored and large others are gray and small

        Parameters
        ----------
        pdb_name: str
        pdb_save_name: str
        network: networkx.Graph
        high_betweenness_df: pd.DataFrame
        save: bool
        """
        highlight_residue_nodes, scores = self._find_high_betweenness_residues(high_betweenness_df, node_keys_to_id_map)

        fig = go.Figure()

        xe_n, ye_n, ze_n = [], [], []
        xe_s, ye_s, ze_s = [], [], []

        for u, v in network.edges():
            x0, y0, z0 = network.nodes[u].get(X, 0), network.nodes[u].get(Y, 0), network.nodes[u].get(Z, 0)
            x1, y1, z1 = network.nodes[v].get(X, 0), network.nodes[v].get(Y, 0), network.nodes[v].get(Z, 0)

            if u in highlight_residue_nodes and v in highlight_residue_nodes:
                xe_s += [x0, x1, None]
                ye_s += [y0, y1, None]
                ze_s += [z0, z1, None]
            else:
                xe_n += [x0, x1, None]
                ye_n += [y0, y1, None]
                ze_n += [z0, z1, None]

        fig.add_trace(go.Scatter3d(
            x=xe_n, y=ye_n, z=ze_n,
            mode='lines',
            line=dict(color='#9ca3af', width=1),
            opacity=0.4, name='Edges',
            showlegend=False,
            hoverinfo='none'
        ))

        fig.add_trace(go.Scatter3d(
            x=xe_s, y=ye_s, z=ze_s,
            mode='lines',
            line=dict(color='#374151', width=4),
            opacity=0.7,
            name='Hub edges',
            showlegend=False,
            hoverinfo='none'
        ))

        normal_nodes = [n for n in network.nodes() if n not in highlight_residue_nodes]
        if normal_nodes:
            fig.add_trace(go.Scatter3d(
                x=[network.nodes[n].get(X, 0) for n in normal_nodes],
                y=[network.nodes[n].get(Y, 0) for n in normal_nodes],
                z=[network.nodes[n].get(Z, 0) for n in normal_nodes],
                mode='markers',
                marker=dict(size=4, color='#4b5563', opacity=0.6, line=dict(width=0.5, color='#374151')),
                name='Other Residues',
                hoverinfo='text',
                text=[self._format_centrality_node_label(self._format_node_label_by_node_id(network, n), splitter=":")
                      for n in
                      normal_nodes]
            ))

        highlight_list = list(highlight_residue_nodes)
        if highlight_list:
            h_scores = [scores.get(n, 0) for n in highlight_list]
            fig.add_trace(go.Scatter3d(
                x=[network.nodes[n].get(X, 0) for n in highlight_list],
                y=[network.nodes[n].get(Y, 0) for n in highlight_list],
                z=[network.nodes[n].get(Z, 0) for n in highlight_list],
                mode='markers+text',
                marker=dict(
                    size=8,
                    color=h_scores,
                    colorscale=[[0, '#0761e8'], [1, '#b91c1c']],
                    colorbar=dict(
                        title="",
                        thickness=20,
                        len=0.75,
                        x=1.02,
                        ticks="outside",
                        tickwidth=2,
                        ticklen=6,
                        tickcolor="gray",
                        tickfont=dict(size=16, color="black", family=FONT_FAMILY)
                    ),
                    line=dict(width=1, color='#374151'),
                    opacity=0.9,
                ),
                text=[self._format_centrality_node_label(self._format_node_label_by_node_id(network, n), splitter=":")
                      for n in
                      highlight_list],
                textposition='top center',
                textfont=dict(size=14, color='black', family=FONT_FAMILY),
                name='High Betweenness Residues',
                hoverinfo='text',
                hovertext=[
                    f"{self._format_centrality_node_label(self._format_node_label_by_node_id(network, n), splitter=':')}"
                    f"<br>Betweenness: {scores.get(n, 0):.6f}"
                    for n in highlight_list
                ],
            ))

        fig.add_annotation(
            x=1.15,
            y=0.5,
            xref="paper",
            yref="paper",
            text=f"Betweenness Centrality",
            textangle=90,
            showarrow=False,
            font=dict(size=20, family=FONT_FAMILY, color="black"),
            align="center"
        )

        fig.update_layout(
            autosize=True,
            title=dict(text=f'{pdb_name} — High Betweenness Centrality Residues',
                       x=0.5, xanchor='center', font=dict(size=_HTML_3D_SIZE)),
            scene=dict(
                domain=dict(x=[0, 1], y=[0, 1]),
                aspectmode="data",
                xaxis=dict(
                    title=dict(text=X.capitalize(), font=dict(size=14)),
                    tickfont=dict(size=14),
                    gridcolor='#e5e7eb'
                ),
                yaxis=dict(
                    title=dict(text=Y.capitalize(), font=dict(size=14)),
                    tickfont=dict(size=14),
                    gridcolor='#e5e7eb'
                ),
                zaxis=dict(
                    title=dict(text=Z.capitalize(), font=dict(size=14)),
                    tickfont=dict(size=14),
                    gridcolor='#e5e7eb'
                ),
                bgcolor='white',
            ),
            height=900,
            width=1200,
            showlegend=True,
            template='plotly_white',
            legend=dict(font=dict(size=_HTML_3D_SIZE)),
            font=dict(family=FONT_FAMILY)
        )

        if save:
            output_file = self.output_path / f'{pdb_save_name}_high_betweenness_network_3d.html'
            fig.write_html(output_file, config=_ZOOM_CONFIG)
            logging.info(f"{output_file} has been saved!")


def parse_residue_pairs(pairs_list):
    if not pairs_list:
        return None
    result = []
    for pair_str in pairs_list:
        left, right = pair_str.split(";")

        def parse_one(r):
            parts = r.split(":")
            chain_id = parts[0]
            residue_number = int(parts[1])
            insertion = parts[2] if len(parts) > 2 else ""
            segment_id = parts[3] if len(parts) > 3 else ""
            return chain_id, residue_number, insertion, segment_id

        result.append((parse_one(left), parse_one(right)))
    return result


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="rinpy compare",
        description=(
            "RinPy NetworkComparator performs comparative analysis "
            "between two protein residue interaction networks (RINs)."
        ),
        epilog=(
            "Example:\n"
            "rinpy compare --source_input_path ./4obe "
            "--target_input_path ./5v9u "
            "--output_path ./analysis_output "
            "--num_modes 20\n\n"
            "# num_modes specifies the number of non-zero Laplacian modes used."
        ),
        formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "--source_input_path",
        required=True,
        help="Directory containing the source state RIN results (e.g., apo structure)."
    )

    parser.add_argument(
        "--target_input_path",
        required=True,
        help="Directory containing the target state RIN results (e.g., ligand-bound structure)."
    )

    parser.add_argument(
        "--output_path",
        required=True,
        help="Directory where comparison results will be saved."
    )

    parser.add_argument(
        "--residue_pairs",
        nargs="+",
        default=None,
        help=(
            "Optional residue pairs for Communication Path Efficiency analysis, "
            "including sequential path efficiency, end-to-end efficiency, "
            "path-restricted internal efficiency, and allosteric coupling.\n\n"
            "Format for each pair:\n"
            "chain:residue_number[:insertion[:segment_id]];"
            "chain:residue_number[:insertion[:segment_id]]\n\n"
            "If insertion or segment_id is not provided, it is treated as an empty string.\n\n"
            "Examples:\n"
            "--residue_pairs \"A:10;A:144\"\n"
            "--residue_pairs \"A:10:A;A:144\"\n"
            "--residue_pairs \"A:10;A:144\" \"A:15;A:30\""
        )
    )

    parser.add_argument(
        "--num_modes",
        type=int,
        default=20,
        help=(
            "Number of non-zero Laplacian eigenvalues used in community structure analysis. "
            "The trivial zero eigenvalue is excluded."
        )
    )

    return parser


def main(argv=None):
    parser = build_parser()
    args = parser.parse_args(argv)
    residue_pairs = parse_residue_pairs(args.residue_pairs)
    logging.info(f"residue_pairs: {residue_pairs}")
    comparator = NetworkComparator(
        source_input_path=args.source_input_path,
        target_input_path=args.target_input_path,
        output_path=args.output_path,
        residue_pairs=residue_pairs
    )
    comparator.run(num_modes=args.num_modes)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os
import time
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.offline as pyo
from scipy.sparse.linalg import eigsh

from rinpy import log_util
from rinpy import utils
from rinpy.constants import RESIDUE_NUMBER, MODE, CHAIN_ID, INSERTION, HIGH_PERCENTAGE_TEMPLATE, RESIDUE_NAME, \
    CENTRALITY_SCORE, TXT_EXT, PDB_EXT, PML_EXT, CSV_EXT, HTML_EXT, PNG_EXT, X, Y, Z, RESIDUE_INDEX, SEGMENT_ID
from rinpy.log_util import log_details
from rinpy.pymol_utils import PymolUtils
from rinpy.style_config import FONT_STYLES, FONT_FAMILY, EDGE_COLOR, COLOR_PALETTE
from rinpy.utils import CentralityType

_HINGE_MODES_DIR = "hinge_modes"


class HingeAnalyzer:
    """
       Performing graph spectral analysis for the generated network of a protein structure to obtain hinge residues
       of each mode (each eigenvector is referred to as a mode). The eigenvector is obtained from Laplacian matrix.

       Parameters
       ----------
       graph
           The protein structure represented as a network, where nodes are residues
           and edges represent local strength interaction or affinity between residues.
       pdb_name
           The identifier of the PDB file being analyzed.
       destination_output_path
           The path where analysis results, plots, or output files will be saved.
       actual_residue_number_map
           contains defined parameters of a node, such as Chain ID, Residue Number and so on.
    """

    def __init__(self, graph: nx.Graph, pdb_name: str, destination_output_path: str,
                 actual_residue_number_map: dict[int, tuple[str, str, int, str, str, str]]):
        self.graph = graph
        self.pdb_name = pdb_name
        self.destination_output_path = Path(destination_output_path)
        self.actual_residue_number_map = actual_residue_number_map
        mpl.rcParams['font.family'] = 'Times New Roman'
        mpl.rcParams['font.size'] = 12
        self.pymol_utils = PymolUtils(pdb_name=pdb_name,
                                      destination_output_path=destination_output_path,
                                      actual_residue_number_map=actual_residue_number_map)

        utils.create_folder_not_exists(self.destination_output_path / self.pdb_name / _HINGE_MODES_DIR)

    def _compute_laplacian_modes(self, num_modes: int) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray] | None:
        """computes the eigenvalues and eigenvectors of the graph laplacian, then computes spectral nodes
            excluding trivial eigenvector
        """
        nodes = sorted(self.graph.nodes())
        laplacian = nx.laplacian_matrix(self.graph, nodelist=nodes).astype(float)

        if not nx.is_connected(self.graph):
            logging.error("Graph must be connected for spectral partitioning.")
            logging.info("Graph must be connected for spectral partitioning.")
            return

        k = num_modes + 1
        eigen_values, eigen_vectors = eigsh(laplacian, k=k, which="SM")

        idx = np.argsort(eigen_values)
        eigen_values = eigen_values[idx]
        eigen_vectors = eigen_vectors[:, idx]

        fiedler_vector = eigen_vectors[:, 1].real
        spectral_modes = eigen_vectors[:, 1:k]  # Fiedler +  {num_modes-1} more num_modes included

        start = time.time()
        self._plot_fiedler(fiedler_vector=fiedler_vector)
        log_util.log_elapsed_time_detail("plot_fiedler: ", start, time.time())

        return eigen_values, eigen_vectors, fiedler_vector, spectral_modes

    def _is_same_chain(self, current_node_id: int, next_node_id: int) -> bool:
        current_node_data = self.actual_residue_number_map.get(current_node_id)
        next_node_data = self.actual_residue_number_map.get(next_node_id)
        if current_node_data is None or next_node_data is None:
            return False
        return current_node_data[1] == next_node_data[1]  # [1] = chain_id

    def _find_hinge_residues_with_sign(self, mode_data: np.ndarray) -> list[int]:
        node_ids_map = dict(enumerate(sorted(self.graph.nodes())))

        mode_signs = np.sign(mode_data)
        abs_magnitude_diff_signs = np.sign(np.diff(np.abs(mode_data)))

        has_sign_change = np.zeros(len(mode_signs), dtype=bool)

        for i in range(len(mode_signs) - 1):
            if mode_signs[i] != mode_signs[i + 1]:
                if self._is_same_chain(node_ids_map[i], node_ids_map[i + 1]):
                    has_sign_change[i] = True

        for idx, abs_diff_sign in enumerate(abs_magnitude_diff_signs):
            if has_sign_change[idx] and abs_diff_sign < 0:
                has_sign_change[idx] = False
                is_same_chain: bool = self._is_same_chain(node_ids_map[idx], node_ids_map[idx + 1])
                if is_same_chain:
                    has_sign_change[idx + 1] = True

        hinge_nodes: set = {node_ids_map[i] for i, is_hinge in enumerate(has_sign_change) if is_hinge}
        return sorted(hinge_nodes)

    def _save_and_plot_mode_to_file(self, mode_data: np.ndarray, mode_name: str, hinge_residues: list[int]) -> None:
        mode_output_path = os.path.join(self.destination_output_path, self.pdb_name, _HINGE_MODES_DIR)
        utils.create_folder_not_exists(mode_output_path)
        nodes = self.graph.nodes()
        actual_hinge_residues_tuple = self.get_actual_hinge_residues_tuple(nodes=nodes,
                                                                           hinge_residues=hinge_residues)

        full_path = os.path.join(str(mode_output_path), f"{mode_name}_hinge_residues{TXT_EXT}")
        self.write_to_file(actual_hinge_residues=actual_hinge_residues_tuple, full_path=full_path)

        residue_keys = [(nodes[n][RESIDUE_NAME], nodes[n][CHAIN_ID], nodes[n][RESIDUE_NUMBER], nodes[n][INSERTION],
                         nodes[n][SEGMENT_ID]) for
                        n in nodes]

        mode_df = pd.DataFrame({
            RESIDUE_INDEX: residue_keys,
            'mode': mode_data
        })

        residue_to_mode = mode_df.set_index(RESIDUE_INDEX)[MODE].to_dict()

        base_pdb_df = utils.get_base_pdb_df(residue_to=residue_to_mode,
                                            destination_output_path=self.destination_output_path,
                                            pdb_name=self.pdb_name)

        out_filename = os.path.join(str(mode_output_path), f"{mode_name}{PDB_EXT}")
        utils.write_ppdb_to_pdb_file_atom_and_hetatom(ppdb=base_pdb_df, out_filename=out_filename)

        full_path = os.path.join(str(mode_output_path), f"{mode_name}{PML_EXT}")

        high_percentage_residues_df = pd.read_csv(os.path.join(str(self.destination_output_path),
                                                               self.pdb_name,
                                                               f'{self.pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=CentralityType.BET.display_name)}'),
                                                  sep=";",
                                                  header=None,
                                                  names=[RESIDUE_INDEX, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID,
                                                         RESIDUE_NUMBER, INSERTION, SEGMENT_ID])
        self.pymol_utils.export_pymol_script_hinge(
            full_path_to_pdb=out_filename,
            residues=hinge_residues,
            full_path=full_path,
            high_percentage_residues=list(
                zip(
                    high_percentage_residues_df[RESIDUE_NAME],
                    high_percentage_residues_df[CHAIN_ID],
                    high_percentage_residues_df[RESIDUE_NUMBER],
                    high_percentage_residues_df[INSERTION],
                    high_percentage_residues_df[SEGMENT_ID]
                )
            )
        )

    def _save_selected_eigen_vectors(self, spectral_modes: np.ndarray) -> None:
        filtered_data_df = pd.DataFrame(spectral_modes)
        num_columns = filtered_data_df.shape[1]
        headers = [f"mode{i}" for i in range(1, num_columns + 1)]
        filtered_data_df.columns = headers

        residue_ids = [
            f"{self.actual_residue_number_map[i + 1][2]}"
            if (i + 1) in self.actual_residue_number_map else str(i + 1)
            for i in range(filtered_data_df.shape[0])
        ]

        chain_ids = [
            f"{self.actual_residue_number_map[i + 1][1]}"
            if (i + 1) in self.actual_residue_number_map else str(i + 1)
            for i in range(filtered_data_df.shape[0])
        ]

        residue_names = [
            f"{self.actual_residue_number_map[i + 1][0]}"
            if (i + 1) in self.actual_residue_number_map else str(i + 1)
            for i in range(filtered_data_df.shape[0])
        ]

        insertions = [
            f"{self.actual_residue_number_map[i + 1][3]}"
            if (i + 1) in self.actual_residue_number_map and self.actual_residue_number_map[i + 1][3] else "''"
            for i in range(filtered_data_df.shape[0])
        ]

        segment_ids = [
            f"{self.actual_residue_number_map[i + 1][4]}"
            if (i + 1) in self.actual_residue_number_map and self.actual_residue_number_map[i + 1][4] else "''"
            for i in range(filtered_data_df.shape[0])
        ]

        filtered_data_df.insert(0, "index", range(1, filtered_data_df.shape[0] + 1))
        filtered_data_df.insert(1, RESIDUE_NAME, residue_names)
        filtered_data_df.insert(2, CHAIN_ID, chain_ids)
        filtered_data_df.insert(3, RESIDUE_NUMBER, residue_ids)
        filtered_data_df.insert(4, INSERTION, insertions)
        filtered_data_df.insert(5, SEGMENT_ID, segment_ids)

        filtered_data_path = self.destination_output_path / self.pdb_name / _HINGE_MODES_DIR / f'{self.pdb_name}_eigenvectors{CSV_EXT}'
        filtered_data_df.to_csv(filtered_data_path, index=False)

    def compute_hinge_residues_with_sign(self, num_modes: int = None) -> None:
        if num_modes is None:
            num_modes = 4

        eigen_values, eigen_vectors, fiedler_vector, spectral_modes = self._compute_laplacian_modes(
            num_modes=num_modes)

        self._save_selected_eigen_vectors(spectral_modes=spectral_modes)

        for i in range(spectral_modes.shape[1]):
            mode_data = spectral_modes[:, i]
            hinge_residues = self._find_hinge_residues_with_sign(mode_data=mode_data)
            if hinge_residues and len(hinge_residues) > 0:
                mode_data = np.where(mode_data < 0, -1, 1)
                mode_name: str = f'{self.pdb_name}_laplacian_mode_{(i + 1)}'
                self._save_and_plot_mode_to_file(mode_data=mode_data,
                                                 mode_name=mode_name,
                                                 hinge_residues=hinge_residues)

                node_ids = sorted(list(self.graph.nodes()))
                cluster_labels = {node_ids[i]: int(mode_data[i]) for i in range(len(node_ids))}
                full_path = os.path.join(self.destination_output_path, self.pdb_name, _HINGE_MODES_DIR,
                                         f'{mode_name}_hinge_interactive_clusters_3d{HTML_EXT}')
                self.plot_graph_interactive_clusters_3d(self.graph, cluster_labels, hinge_residues, full_path)
                start = time.time()

                full_path = os.path.join(self.destination_output_path, self.pdb_name, _HINGE_MODES_DIR,
                                         f"{mode_name}_graph_clusters_2d{PNG_EXT}")
                self.plot_graph_clusters_2d(self.graph, cluster_labels, hinge_residues, full_path=full_path)
                log_util.log_elapsed_time_detail("plot_graph_clusters_2d: ", start, time.time())

    @staticmethod
    def get_actual_hinge_residues_tuple(nodes, hinge_residues: list[int]) -> list[str]:
        return [(
            f"{nodes[hr][RESIDUE_NAME]};{nodes[hr][CHAIN_ID]};{nodes[hr][RESIDUE_NUMBER]};'{nodes[hr][INSERTION]}';'{nodes[hr][SEGMENT_ID]}'")
            for hr in hinge_residues
        ]

    def _get_full_save_path(self, filename: str, extension: str = "png") -> str:
        return os.path.join(self.destination_output_path, self.pdb_name, f"{self.pdb_name}_{filename}.{extension}")

    def write_to_file(self, actual_hinge_residues, full_path=None):
        if full_path is None:
            full_path = self._get_full_save_path(filename="hinge_residues", extension="txt")

        with open(full_path, 'w') as f:
            for i, hinge_residue in enumerate(actual_hinge_residues):
                if i < len(actual_hinge_residues) - 1:
                    f.write(f"{hinge_residue}\n")
                else:
                    f.write(f"{hinge_residue}")

    @log_details("Plotting Fiedler vector")
    def _plot_fiedler(self, fiedler_vector: np.ndarray, sort: bool = False) -> None:
        """ Plots and optionally saves the Fiedler vector.

        Parameters:
        - fiedler_vector (np.ndarray): The Fiedler vector to plot.
        - save_path (str, optional): Path to save the figure (without extension). If None, just shows the plot.
        - sort (bool): Whether to plot the sorted Fiedler vector (for smoother visualization).
        - file_format (str): Image format for saving (e.g., 'png', 'pdf').
        - dpi (int): Resolution of the saved image.
        """
        full_path = self._get_full_save_path(filename="fiedler_vector")
        vector = fiedler_vector.copy()
        if sort:
            vector = np.sort(vector)

        fig, ax = plt.subplots(figsize=(10, 6))

        next_round = int(np.ceil(len(vector) / 100.0)) * 100
        ax.set_xlim(0, next_round)

        residue_labels = [
            f"{i + 1} ({self.actual_residue_number_map[i + 1][1]}, {utils.get_residue_id_by_tuple(self.actual_residue_number_map[i + 1])})"
            for i in range(len(vector))
            if (i + 1) in self.actual_residue_number_map
        ]
        x_ticks = ax.get_xticks().astype(int)
        x_ticks = x_ticks[x_ticks < len(vector)]

        if (len(vector) - 1) not in x_ticks:
            x_ticks = np.append(x_ticks, len(vector) - 1)

        ax.plot(vector, marker='o', linestyle='-', color='blue', label='Fiedler Vector')
        ax.axhline(y=0, color='red', linestyle='--', linewidth=1, label='Zero Line')
        ax.set_title(f"Fiedler Vector - {self.pdb_name}", fontdict=FONT_STYLES["title"])
        ax.set_xticks(x_ticks)
        ax.set_xticklabels([f"{residue_labels[i]}" for i in x_ticks], rotation=45, **{
            "fontsize": FONT_STYLES["xtick"]["labelsize"],
            "fontfamily": FONT_STYLES["xtick"]["fontfamily"]
        })
        ax.set_xlabel("Residue Index (Chain ID, Residue Number)", fontdict=FONT_STYLES["xlabel"])

        for label in ax.get_yticklabels():
            label.set_fontsize(FONT_STYLES["ytick"]["labelsize"])
            label.set_fontfamily(FONT_STYLES["ytick"]["fontfamily"])

        ax.set_ylabel("Value", fontdict=FONT_STYLES["ylabel"])
        ax.grid(True, linestyle=':', alpha=0.7)
        ax.legend(prop={"family": FONT_FAMILY, "size": FONT_STYLES["legend"]["fontsize"]})

        next_round = int(np.ceil(len(vector) / 100.0)) * 100
        ax.set_xlim(0, next_round)

        fig.tight_layout()

        fig.savefig(full_path, dpi=300)
        logging.info(f"Saving Fiedler vector to: {os.path.abspath(full_path)}")
        plt.close(fig)

    def plot_graph_clusters_2d(self, graph, cluster_labels, hinge_residues, full_path=None):
        if full_path is None:
            full_path = self._get_full_save_path(filename="graph_clusters_2d")
        unique_clusters = sorted(set(cluster_labels.values()))
        color_map = {cluster_id: COLOR_PALETTE[i % len(COLOR_PALETTE)] for i, cluster_id in
                     enumerate(unique_clusters)}

        fig, ax = plt.subplots()

        for cluster_id in unique_clusters:
            cluster_nodes = [n for n in graph.nodes() if cluster_labels[n] == cluster_id]
            xs = [graph.nodes[n]['x'] for n in cluster_nodes]
            ys = [graph.nodes[n]['y'] for n in cluster_nodes]
            ax.scatter(xs, ys, s=100, label=f"Cluster ({cluster_id})", color=color_map[cluster_id],
                       edgecolor=color_map[cluster_id], linewidth=1, alpha=0.6)

            for n in cluster_nodes:
                x, y = graph.nodes[n]['x'], graph.nodes[n]['y']
                ax.text(x, y, str(n), fontsize=6, ha='center', va='center', color='black', alpha=0.6)

        for u, v in graph.edges():
            x = [graph.nodes[u]['x'], graph.nodes[v]['x']]
            y = [graph.nodes[u]['y'], graph.nodes[v]['y']]
            ax.plot(x, y, color=EDGE_COLOR, alpha=0.5, linewidth=0.5)

        hinge_x = [graph.nodes[n]['x'] for n in hinge_residues]
        hinge_y = [graph.nodes[n]['y'] for n in hinge_residues]
        ax.scatter(hinge_x, hinge_y, s=120, c='yellow', edgecolors='black', marker='*',
                   label='Hinge Residues', zorder=10)

        for n in hinge_residues:
            x, y = graph.nodes[n]['x'], graph.nodes[n]['y']
            ax.text(x, y, str(n), fontsize=7, ha='center', va='center', color='black', fontweight='bold')

        ax.set_xlabel("X Coordinate", fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel("Y Coordinate", fontdict=FONT_STYLES["ylabel"])
        ax.set_title("2D Spectral Clustering of Residue Interaction Network", fontdict=FONT_STYLES["title"])
        ax.legend(prop={"family": FONT_FAMILY, "size": FONT_STYLES["legend"]["fontsize"]})
        ax.axis('equal')
        ax.grid(True, which='both', axis='both', color='gray', linestyle='-', linewidth=0.5, alpha=0.3)

        self.update_tick_label_style(ax.get_xticklabels())
        self.update_tick_label_style(ax.get_yticklabels())

        fig.tight_layout()
        fig.savefig(full_path, dpi=300)
        plt.close(fig)

    @staticmethod
    def update_tick_label_style(labels):
        for label in labels:
            label.set_fontsize(FONT_STYLES["ytick"]["labelsize"])
            label.set_fontfamily(FONT_STYLES["ytick"]["fontfamily"])

    def plot_graph_clusters_3d(self, graph, cluster_labels, hinge_residues):
        """ 3D plot (X-Y-Z) with manually assigned high-contrast colors."""
        full_path = self._get_full_save_path(filename="graph_clusters_3d")

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        unique_clusters = sorted(set(cluster_labels.values()))
        color_map = {cluster_id: COLOR_PALETTE[i % len(COLOR_PALETTE)] for i, cluster_id in
                     enumerate(unique_clusters)}

        legend_handles = []

        for cluster_id in unique_clusters:
            cluster_nodes = [n for n in graph.nodes() if cluster_labels[n] == cluster_id]
            xs = [graph.nodes[n]['x'] for n in cluster_nodes]
            ys = [graph.nodes[n]['y'] for n in cluster_nodes]
            zs = [graph.nodes[n]['z'] for n in cluster_nodes]
            scatter = ax.scatter(xs, ys, zs, s=100, label=f"Cluster ({cluster_id})", color=color_map[cluster_id])
            legend_handles.append(scatter)

        for u, v in graph.edges():
            x = [graph.nodes[u][X], graph.nodes[v][X]]
            y = [graph.nodes[u][Y], graph.nodes[v][Y]]
            z = [graph.nodes[u][Z], graph.nodes[v][Z]]
            ax.plot(x, y, z, color=EDGE_COLOR, alpha=0.5, linewidth=0.5)

        hinge_scatter = []
        for n in hinge_residues:
            if n in graph.nodes():
                x, y, z = graph.nodes[n][X], graph.nodes[n][Y], graph.nodes[n][Z]
                ax.scatter(x, y, z, s=150, c='yellow', edgecolors='black', marker='*')
                ax.text(x, y, z, str(n), fontsize=7, ha='center', va='center', color='black', fontweight='bold')
                hinge_scatter.append(ax.scatter([], [], s=150, c='yellow', edgecolors='black', marker='*'))

        legend_handles.append(hinge_scatter[0])
        ax.legend(handles=legend_handles,
                  labels=[f"Cluster ({i})" for i in unique_clusters] + ["Hinge Residues"],
                  prop={"family": FONT_FAMILY, "size": FONT_STYLES["legend"]["fontsize"]}, loc='upper right',
                  bbox_to_anchor=(1.2, 1.01))

        ax.set_xlabel("X Coordinate", fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel("Y Coordinate", fontdict=FONT_STYLES["ylabel"])
        ax.set_zlabel("Z Coordinate", fontdict=FONT_STYLES["zlabel"])
        plt.title("3D Spectral Clustering of Residue Interaction Network", fontdict=FONT_STYLES["title"])

        self.update_tick_label_style(ax.get_xticklabels())
        self.update_tick_label_style(ax.get_yticklabels())
        self.update_tick_label_style(ax.get_zticklabels())

        plt.tight_layout()
        fig.savefig(full_path, dpi=300)
        plt.close(fig)

    # @staticmethod
    @staticmethod
    def plot_graph_interactive_clusters_3d(graph, cluster_labels, hinge_residues, full_path):
        unique_clusters = sorted(set(cluster_labels.values()))
        color_map = {
            cluster_id: COLOR_PALETTE[1] if cluster_id < 0 else COLOR_PALETTE[0]
            for cluster_id in unique_clusters
        }
        node_traces = []
        for cluster_id in unique_clusters:
            cluster_nodes = [n for n in graph.nodes if cluster_labels[n] == cluster_id]
            x = [graph.nodes[n][X] for n in cluster_nodes]
            y = [graph.nodes[n][Y] for n in cluster_nodes]
            z = [graph.nodes[n][Z] for n in cluster_nodes]

            labels = []
            for n in cluster_nodes:
                node = graph.nodes[n]
                seg_id = node.get(SEGMENT_ID)
                seg_text = f", Segment ID: {seg_id}" if seg_id not in [None, '', "''"] else ''
                label = (
                    f"{node[RESIDUE_NAME]} "
                    f"(Residue Number: {utils.get_residue_id(node)}, Chain: {node[CHAIN_ID]}{seg_text}), "
                    f"Cluster ({cluster_id})"
                )
                labels.append(label)

            trace = go.Scatter3d(
                x=x,
                y=y,
                z=z,
                mode='markers',
                marker=dict(size=5, color=color_map[cluster_id]),
                text=labels,
                hoverinfo='text',
                name=f'Cluster {cluster_id}',
                hoverlabel=dict(
                    font=dict(color='white')
                )
            )
            node_traces.append(trace)

        edge_x, edge_y, edge_z = [], [], []
        for u, v in graph.edges():
            edge_x += [graph.nodes[u][X], graph.nodes[v][X], None]
            edge_y += [graph.nodes[u][Y], graph.nodes[v][Y], None]
            edge_z += [graph.nodes[u][Z], graph.nodes[v][Z], None]

        edge_trace = go.Scatter3d(
            x=edge_x,
            y=edge_y,
            z=edge_z,
            mode='lines',
            line=dict(color='#1f4e79', width=1),
            hoverinfo='none',
            name='Edges'
        )

        hinge_trace = go.Scatter3d(
            x=[graph.nodes[n][X] for n in hinge_residues],
            y=[graph.nodes[n][Y] for n in hinge_residues],
            z=[graph.nodes[n][Z] for n in hinge_residues],
            mode='markers+text',
            marker=dict(size=7, color='yellow', symbol='diamond', line=dict(color='black', width=1)),
            name='Hinge Residues',
            hoverinfo='text',
            hovertext=[
                f"<b>Residue Name: {graph.nodes[n][RESIDUE_NAME]}</b><br>"
                f"<b>Chain: {graph.nodes[n][CHAIN_ID]}</b><br>"
                f"<b>Hinge Residue Number: {utils.get_residue_id(graph.nodes[n])}</b>"
                + (f"<b>Segment ID</b>: {graph.nodes[n][SEGMENT_ID]}<br>" if graph.nodes[n][SEGMENT_ID] not in
                                                                             [None, '', "''"] else "")
                + f"<br>────────────────────<br>"
                  f"x={graph.nodes[n]['x']:.3f}<br>"
                  f"y={graph.nodes[n]['y']:.3f}<br>"
                  f"z={graph.nodes[n]['z']:.3f}"
                for n in hinge_residues],
            hoverlabel=dict(
                bgcolor='yellow',
                font=dict(color='black')
            )
        )

        fig = go.Figure(data=[edge_trace, hinge_trace] + node_traces)

        axis_style = dict(
            title_font=dict(family=FONT_FAMILY, size=18),
            tickfont=dict(family=FONT_FAMILY, size=16)
        )
        fig.update_layout(
            scene=dict(
                xaxis=dict(title="X Coordinate", **axis_style),
                yaxis=dict(title="Y Coordinate", **axis_style),
                zaxis=dict(title="Z Coordinate", **axis_style),
            ),
            margin=dict(l=0, r=0, b=0, t=40),
            title=dict(
                text="Interactive 3D Spectral Clustering of Residue Interaction Network",
                font=dict(family="Times New Roman", size=24),
                x=0.5,
                xanchor="center",
                y=0.95,
                yanchor="top"
            ),
            template='plotly',
            legend=dict(
                font=dict(family=FONT_FAMILY, size=16),
                title=dict(
                    font=dict(family=FONT_FAMILY, size=16)
                )
            ),
            scene_camera=dict(
                eye=dict(x=1.2, y=1.2, z=1.2)
            )
        )

        pyo.plot(fig, filename=full_path, auto_open=False)
        logging.info(f"Interactive plot saved to: {full_path}")


def main():
    pass


if __name__ == '__main__':
    main()
